#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <bio.h>

#include "httpd.h"
#include "httpsrv.h"
#include "misc.h"

static 	int	notefun(void *a, char *msg);
static 	int	execsafely(HConnect *c, char *path, char **argv, int hctl);
static 	int	nomount(HConnect *c);
static 	int	getresheader(HConnect *c, char *status, char **resheader);


int
execute(HConnect *c, char *hargv[], int hctl)
{
	char *p, *q;
	char *params;
	int k, n, hargc;
	char **argv;
	char *path;

	params = c->req.params;
	/*	params points "alice;bob"
	*	In creating argv for exec call,
	*	we should allocate memory for argv.
	*	argv[0] must be reserved for basename of exec file
	*	argv[n-1] must be nil where n is num of elements
	*	so, I use ";alice;bob;" to apply getfield.
	*	-Kenar-		*/

	//logit(c, "DEBUG: params %s",params);

	if(params == nil)
		argv = hargv;
	else{
		for(n = 0; hargv[n]; n++);
		hargc = n;
			
		n = strlen(";") + strlen(params) + 1 + UTFmax;
		q = halloc(c, n);
		snprint(q, n, "%s;", params);
		params = q;
		n = 0;
		for(p = params; *p; p++)
			if(*p == ';') n++;
		argv = halloc(c, (hargc + n+1) * sizeof(char*));
		n = getfields(params, &argv[hargc], n+1, 0, ";");
		for(k = 0; k < hargc; k++)
			argv[k] = hargv[k];
		/* don't forget to hurlunesc */
		for(k = hargc; k < hargc + n -1; k++)
			argv[k] = hurlunesc(c, argv[k]);
		/* it is not good idea to trim extra spaces */
		argv[hargc + n-1] = nil;
	}
	path = argv[0];
	p = strrchr(argv[0],'/');
	argv[0] = p + 1;

	/*	if it takes so much time to complete, kill the process
	*	-Kenar-	*/
	return execsafely(c, path, argv, hctl);
}


extern int mntok;
extern int authed;
extern int cgitimeout;
static int notestatus; // alarm note to CGI is received
static int timeoutreset;
static int cpid;

/*	This code intends to get exit status from the child	*/
static int
execsafely(HConnect *c, char *path, char **argv, int hctl)
{
	int fd[2], fd0, fd1;
	long n;//,m;
	Biobuf ib;
	int status;
	char msg[ERRMAX];
	char *exitmsg;
	//char *resheader;
	Hio *hout;
	char *p, buf[4096];


	/* 	wait child to finish.
	*	If it takes too much time, kill the child and get message  */
	if(pipe(fd)<0)
		hfail(c, HInternal);

	switch(cpid = rfork(RFFDG|RFPROC|RFNOTEG|RFREND|RFNAMEG)){ /* assign = */
	/*	RFNAMEG: the new process inherits a copy of the parent's name space.
	*	RFNAMEG is required so that nomnt() does not make effect to parent
	*	namespace	*/
	case -1:
		logit(c, "error: fork # %r");
		hfail(c, HInternal);
		return -1;
	case 0: // child: read stdin, and write to stdout
		dup(fd[1], 1); // the stdout is redirected to fd
		dup(fd[1], 2); // the stderr is also redirected to fd
		close(fd[1]);
		if(!authed)
			hbind(c, "/mnt", "/etc", MREPL);
		if(!mntok)
			nomount(c);
		exec(path, argv);
		/* not reached in normal exec */
		print("500 Internal Server Error");
		logit(c,"error: exec %s # %r", path);
		exits("500 Internal Server Error");// don't return
		break;
	default:// parent: read data from the child
		close(fd[1]);
		break;
	}
	atnotify(notefun,1);
	alarm(cgitimeout*1000);
	notestatus = 0;

	fd0 = fd[0];
	hout = &c->hout;

	status = 0;
	if(hctl == HCFull){
		char *p;
		int n;

		/*	timeout reset might not work well
		*	timeout reset before header output might be safe
		*	-Kenar- 	*/
		Binit(&ib, fd0, OREAD);
		while(((p = Brdstr(&ib,'\n',1)) && *p) || timeoutreset){
			timeoutreset = 0;
			if(p == nil || *p == 0)
				continue;
			n = strlen(p); // then n > 0 because *p != 0
			if(p[n - 1] == '\r'){
				p[n - 1] == 0;
				if(n == 1)
					break;
			}
			hprint(hout, "%s\r\n", p);
		}
		hprint(hout,"\r\n");
		while((n = Bread(&ib, buf, sizeof(buf))) > 0 || timeoutreset){
			timeoutreset = 0;
			if(n <= 0)
				continue;
			hwrite(hout, buf, n);
		}
		logit(c, "done %s", path);
		return hflush(hout);
	}
	if(hctl == HCPlus){
		char *av[2],*mt;
		int m;

		/* mt is mymetype in handler */
		mt = db_get(cgih,"Content-Type:");
		if(mt == nil)
			mt = "";

		/*	timeout reset might not work well
		*	timeout reset before header output might be safe
		*	-Kenar- 	*/
		Binit(&ib, fd0, OREAD);
		while(((p = Brdstr(&ib,'\n',1)) && *p) || timeoutreset){
			timeoutreset = 0;
			if(p == nil || *p == 0)
				continue;
			n = strlen(p);
			if(n > 0 && p[n-1] == '\r'){
				n--;
				p[n] = 0;
			}
			if(n == 0)
				break;

			if(strncmp("#DEBUG:",p,7) == 0){
				logit(c,"%s", p);
				continue;
			}

			n = getfields(p,av,2,1," ");
			if(n < 1){
				logit(c, "# error: CGI format");
				return hfail(c, HInternal);
			}
			if(n > 1)
				av[1] = trim(av[1]);
			else
				av[1] = "";
			if(strcmp("Content-Length:",av[0]) != 0)
				db_put(&cgih,av[0],av[1]);
		}

		fd1 = create("/tmp/...", OWRITE, 0600);
		if(fd1 < 0)
			return hfail(c, HInternal);
		/*	we cannot use "read()" here, because some data might be left in
		 *	input buffer ib
		 *	incoming data comes from CGI, the connection will be closed in finish.	*/
		m = 0;
		while((n = Bread(&ib, buf, sizeof(buf))) > 0 || timeoutreset){
			timeoutreset = 0;
			if(n <= 0)
				continue;
			write(fd1, buf, n);
			m += n;
		}
		close(fd1);

		//logit(c,"CGI: body size=%d", m);

		if(strcmp(mt,"text/html") == 0){
			if(m == 0){
				p = db_get(cgih,"Status:");
				if(p && strncmp(p,"200 ", 4) != 0 && strncmp(p,"301 ", 4) != 0)
					return hfail0(c, p, nil, nil);
			}
			p = db_get(cgih,"Location:");
			if(p)
				return hmoved(c, p, nil);
		}
	}
	else if(hctl == HCNo){
		fd1 = create("/tmp/...", OWRITE, 0600);
		if(fd1 < 0)
			return hfail(c, HInternal);
		while((n = read(fd0,buf,sizeof(buf))) > 0 || timeoutreset){
			timeoutreset = 0;
			if(n <= 0)
				continue;
			write(fd1, buf, n);
		}
		close(fd1);
		status = hflush(hout);
	}
	else{
		logit(c,"# invalid hctl");
		return hfail(c, HInternal);
	}

	status = waitfor(cpid, msg);
	alarm(0);
	atnotify(notefun,0);
	close(fd0);
	if(notestatus){
		logit(c,"alarm done to pid=%d", cpid);
		return hfail(c, HCgiTimeout);
	}

	/*	example of msg:
	*	httpd 1653: 500 Internal Server Error
	*	therefore we must extract what we want	*/
	exitmsg = nil;
	if(msg[0] != 0){
		exitmsg = strchr(msg, ':');
		if(exitmsg && exitmsg[1] == ' ')
			exitmsg += 2;
		else
			exitmsg = msg;
		//logit(c,"DEBUG: child %d exited with msg [%s]", cpid, exitmsg);

		/*** the following codes make the debug difficult!
		if(!httpstatus(exitmsg))
			return hfail(c,HCGIErr,exitmsg);
		c->head.closeit = getresheader(c, exitmsg, &resheader);
		if(strcmp(resheader,"200 OK") != 0)
			return hfail0(c,exitmsg,nil,nil);
		***/
	}

	p = db_get(cgih, "Connection:");
	if(p){
		if(cistrcmp(p,"close") == 0)
			c->head.closeit = 1;
		db_del(&cgih, "Connection:");
	}

	p = db_get(cgih, "X-CGI-Pass:");
	if(p){
		if(*p == 0){
			int n,m;
			n = strlen(c->req.path);
			m = n + 20;
			p = halloc(c,m);
			if(c->req.path[n-1] == '/')
				snprint(p,m,"/doc%sindex.html",c->req.path);
			else
				snprint(p,m,"/doc%s",c->req.path);
		}
		//logit(c,"DEBUG: pass %s",p);
		p = hstrdup(c, p);
		db_del(&cgih, "X-CGI-Pass:");
		status = sendit(c, p, hctl);
	}
	else
		status = sendit(c, "/tmp/...", hctl);

	if(exitmsg)
		logit(c, "done %s # [%s]",path, exitmsg);
	else
		logit(c, "done %s", path);
	return status;
}

static int
nomount(HConnect *c)
{
	int status;
	status = rfork(RFNOMNT|RFENVG);
	if(status < 0){
		logit(c,"error: rfork nomnt # %r");
		return hfail(c, HInternal);
	}
	return 0;
}


/*	"/proc" is required for postnote	*/

static int
notefun(void *a, char *msg)
{
	USED(a);
	if(strncmp(msg, "timeout ",8) == 0){
		/*	alarm reset. note that alarm is in progress
		*	the msg format is: timeout n
		*	where n is a digit in unit of second	*/
		msg = strchr(msg,' ');
		alarm(atoi(msg)*1000);
		//syslog(0,"http","DEBUG: alarm is reset to %s", msg);
		timeoutreset = 1;
		return 1;
	}
 	postnote(PNGROUP, cpid, msg);
	if(strcmp(msg, "alarm") == 0){
		notestatus = 1;
		return 1;
	}
	return 0;
}


/*
*	we have two cases of uri
*	scheme://host/foo
*	/foo		# where should I go?
*
*	Note: in last case, hmoved() cares for putting scheme://host
*	but the original hmoved() can care only http.
*	must be fixed!
*	scheme should be kept in redirection.
*
*	It's better to close connection if destination is not
*	this name space.
*	uri:
*	"/~who/path"
*	"/path"
*/
char *
addoptions(HConnect *c, char *uri)
{
	char *new;
	int n;
	char *p, *u;
	char *option;
	option = c->req.options;
	if(option == nil)
		option = "";
	u = nil; // user in uri
	if(uri[0] == '/'){
		if(uri[1] == '~'){
			p = strchr(uri+2, '/');
			if(p)
				*p = 0;
			u = hstrdup(c, uri+2);
			if(p)
				*p = '/';
		}
		if(xstrcmp(c->req.user, u) != 0)
			c->head.closeit = 1;
	}
	else
		c->head.closeit = 1;
	n = strlen(uri) + strlen(option) + 1 + UTFmax;
	new = halloc(c, n);
	snprint(new, n, "%s%s", uri, option);
	return new;
}

int
httpstatus(char *status)
{
	if(strlen(status) > 3
	&& isdigit(status[0]) &&  isdigit(status[1]) &&  isdigit(status[2])
	&& status[3] == ' ')
		return 1;
	return 0;
}


/*
*	getresheaders() get HTTP headers.
*	status is exit-message of a CGI program
*	If stasus is nil or "" , it means normal exit.
*	CGI program should end with exit-message of some HTTP
*	response rule if it does not end normally.
*	However my old CGI programs are coded in loose manner;
*	they exits with groundless message.
*	Therefore getresheader should handle these exit-message
*	so that it is compatible with older version of Pegasus:
*	if status is not HTTP response format we should assign
*		status = "200 OK"
*		closeit = 1
*/
static int
getresheader(HConnect *c, char *status, char **resheader)
{
	char *p;
	int f=0; // f = 1 is status is HTTP response message
	int closeit = 0;
	if(status == nil || strlen(status) < 3){
		*resheader = "200 OK";
		return 0;
	}
	if(httpstatus(status))
		f = 1;
	if(strlen(status) > 3 && !isspace(status[3]))
		f = 0;
	if(f == 0){
		status = "500 Internal Error";
		closeit = 1;
	}
	else{
		status = hstrdup(c, status);
		p = strchr(status, '#');
		if(p){
			*p++ = 0;
			p = trim(p);
			if(cistrcmp(p, "close") == 0)
				closeit = 1;
			else if(cistrcmp(p, "keep") == 0)
				closeit = 0;
			status = trim(status);
		}
		else{
			if(status[0] == '1' || status[0] == '2')
				closeit = 0;
			else
				closeit = 1;
		}
	}
	*resheader = status;
	return closeit;
}

