#include <u.h>
#include <libc.h>
#include <bio.h>

#include "httpd.h"
#include "httpsrv.h"
#include "misc.h"

/*	exec handler
*
*	# data format:
*	path	mimetype	hctl	path	args ...
*
*	# example:
*	*.cgi	text/html	+	$target
*	/dav/*	text/html	+	/bin/davcgi
*	/cgi-bin/foo.cgi/*	text/html + /doc/cgi-bin/foo.cgi
*
*	# return values
*	nil: need not to handle
*	else: handling info
*	-Kenar-	*/
Handle*
handler(HConnect *c, char *file, char *path)
{
	char *line, *s, *p, *syspath;
	Biobuf *b;
	char *field[16+1], **v;
	Handle *hp;
	int i, n, ok;
	if(!path)
		return nil;
	b = Bopen(file, OREAD);
	if(b == nil)
		return nil;

	syspath = hcat2(c, "/doc", path);

	ok = 0;
	p = halloc(c, 256);
	while((line = Brdline(b, '\n')) != nil){
		line[Blinelen(b)-1] = 0;
		//logit(c,"DBG: line: %s",line);
		s = strchr(line, '#');
		if(s != nil)
			*s = '\0'; 	/* chop comment */
		strncpy(p, line, 256);
		n = tokenize(p, field, 16);
		if(n < 4)
			continue;
		if(match(path, field[0])){
			ok = 1;
			field[n] = nil;
			break;
		}
	}
	Bterm(b);
	if(!ok)
		return nil;

	/*	then field[0] field[1] ... are allocated memory	*/

	/*
	for(i = 0; i < n; i++)
		logit(c,"DBG: field[%d]=%s", i, field[i]);
	*/


	hp = halloc(c, sizeof(Handle));

	hp->mimetype = field[1];
	
	switch(*field[2]){
	case '0': hp->hctl = HCNo;
		break;
	case '+': hp->hctl = HCPlus;
		break;
	case '1': hp->hctl = HCFull;
		break;
	case '*': hp->hctl = HCWild;
		break;
	default: hp->hctl = HCNo;
	}

	/* we must have a room for additional $target at the end */
	v = halloc(c, n * sizeof(char*));
	for(n = 3; field[n]; n++)
		v[n - 3] = field[n];
	v[n -3] = nil;

	/*	replace $target by syspath
	*	syspath is "/doc" prepended string of the path	*/
	for(n = 0; v[n]; n++)
		if(strcmp(v[n],"$target") == 0){
			v[n] = syspath;
		}
	p = v[0];
	if(p[0] != '/' || p[strlen(p)-1] == '/'){
		logit(c, "#error: handler: %s", p);
		return nil;
	}
	hp->argv = v;

	/*	Example of CGI/1.1 environment variable
	*	We first consider the case
	*		http://host/foo/bar
	*	where foo and bar may contain slashes.
	*	CGI/1.1(RFC3875) supposes that CGI script is a part of the URI,
	*	say, "/foo".
	*	Interpreting to our variable,
	*	intpath is "/doc/foo/bar" and v[0] is "/doc/foo"
	*	Then the RFC requires
	*		PATH_INFO = "/bar"
	*		PATH_TRANSLATED = "/doc/bar"
	*		SCRIPT_NAME = "/foo"
	*	We must keep the relation:
	*		URIPATH == SCRIPT_NAME + PATH_INFO    --- (1)
	*
	*	Suppose that intpath is "/doc/foo/bar" but v[0] is "/bin/baz" or "/doc/baz".
	*	The case is not supposed by CGI/1.1.
	*	Then we might be allowed to assign anything keeping the relation (1).
	*	For example,
	*		PATH_INFO is "/foo/bar"
	*		PATH_FRANSLATED is "/doc/foo/bar"
	*		SCRIPT_NAME is ""
	*
	*
	*	Next, we should consider regular users case.
	*		http://host/~alice/foo/bar
	*	then intpath is "/doc/foo/bar". Under the condition
	*	if v[0] is "/doc/foo" then 
	*		PATH_INFO = "/bar"
	*		PATH_TRANSLATED = "/doc/bar"
	*		SCRIPT_NAME = "/~alice/foo"
	*	if v[0] is "/bin/baz" then
	*		PATH_INFO = "/~alice/foo/bar"
	*		PATH_FRANSLATED = "/doc/foo/bar"
	*		SCRIPT_NAME = ""
	*	or the following convention also might be possible
	*		PATH_INFO = "/foo/bar"
	*		PATH_FRANSLATED = "/doc/foo/bar"
	*		SCRIPT_NAME = "/~alice"
	*
	*
	*	We temporally assign in case v[0] == "/doc/foo"
	*		pathinfo = "/bar"
	*		scriptname = "/foo"
	*	and in case v[0] == "/bin/baz"
	*		pathinfo = "/foo/bar"
	*		scriptname = nil
	*	for both hosts and regular users.
	*	They must be converted to legal environment variables
	*	accommodate to CGI/1.1.
	*/

	/*	assign scriptname and pathinfo	*/
	/*	we can assume v[0] does not end with '/'	*/
	n = deeper(syspath,v[0]);
	if(n){
		/*	case v[0] that CGI/1.1 supposes
		*	if v[0] == "/doc/foo" and syspath == "/doc/foo/bar"
		*	then n == 8	*/
//		hp->pathinfo = syspath + n;
		hp->pathinfo = c->req.path + n - 4;
		hp->scriptname = v[0] + 4; /* skip first "/doc" */
	}
	else{
//		hp->pathinfo = syspath + 4;
		hp->pathinfo = c->req.path;
		hp->scriptname = nil;
	}
	return hp;
}

