#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <auth.h>
#include <mp.h>
#include <libsec.h>
#include <bio.h>
#include <ndb.h>

#include "httpd.h"
#include "httpsrv.h"
#include "hauth.h"
#include "misc.h"
#include "okip.h"

const char *serversoft="Pegasus/2.7c (Plan9)";
const char *httpdconf = "/sys/lib/httpd.conf";

const char *HTTPLOG = "http";
const char *netdir = "/net";

#define setenv(var, value) value?putenv(var, value):putenv(var, "")

/*	default options	*/
static char *webbase = "/usr/web";
static char *namespace = "/lib/namespace.httpd";
static char *rewrite = "/sys/lib/httpd.rewrite"; // system rewrite-file
char *port = "80"; // replaced later by really accepted port. used in log.c
static int	server = 0; // 0: runs under the control of listen. 1: runs in server mode.
int mntok=0;

static char *runas="none";
static int usermode = 0; // 1: httpd service by the user who invoked httpd
static int monmode = 0; // 1: not run as daemon; monitored by `mon'
static int noweb = 0; // 0: allows $home/web as HomePage if present

/*	the followings may be required tuning for your server
*	and can be reset in configuration file "/sys/lib/httpd.conf"	*/
char *charset = nil;  /* HTTP char set. used in send.c */
static int parsetimeout = 15; // timeout to parse header (in unit of sec)
static int waittimeout1 = 15; // wait timeout for a non-authenticated client (in unit of sec)
static int waittimeout2 = 900; // wait timeout for an authenticated client (in unit of sec)
int cgitimeout = 5; // timeout for CGI (in unit of sec)
static int posttimeout = 15*60; // timeout for post data (in unit of sec)
/* maxpost1, maxpost2: post data size limit
*  that must less than 512MB(comes from ramfs limit)	*/
static int maxpost1 = 10; // for non-authenticated clients (in unit of MB)
static int maxpost2 = 100; // for authenticated clients (in unit of MB)
static int maxconnect = 50; // max tcp connection by the IP
int obstime = 3; // observation time to detect burst access (in unit of sec)
int lockouttime = 60*3; // lockout time for maxconnect and maxaccess (in unit of sec)
static int contmax = 100; // max persistent continuation count for safty

/*	basic constant that are given later. they are for real host	*/
static char myip[16]; //  server IP. 202.250.160.40 for example
char *scheme;
/*	scheme is determined by reading connected port number
*	the value is "http" or "https"
*	and is used in constructing uriroot
*	and local redirection 	*/
char *uriroot;
/*	uriroot is
*		"http://ar.aichi-u.ac.jp:8080"
*		"https:192.168.1.2"
*	for example.
*	uriroot is not given explicitly by the request.
*	Therefore we must guess how our client has accessed to our host
*	and we send the info to our client in local redirection.
*	The value will be assigned in httpd.c and then
*	default port will be extracted.
*/


static char *hostdomain;
/*	hostdomain: full domain name of the host, either for real, virtual host	*/

static char *hostroot;
static char *hroot;
/* hostroot: for example /usr/bob/www; real host's webroot
*  hroot: (=httpd root): for example /usr/bob/www, /usr/alice/web; equiv. $web in the document.
*  hostroot is required because real host goves some help to users
*/


/* used internally for control */

static int waittimeout;
static int mnt; // namespace configuration is finished
static int cont; // current persistent connection continuation count
static char *webmnt; // to be "/usr/web/mnt"
static int binding; // 0: main document, 1: virtual host doc, 2: user's doc
static char *certificate;
static uchar *cert;
static int certlen;
static int ntossed;
/* 	ntossed: current tossed count that controls redirection loop
*	in persitent connection. */
Db *cgih;
/* 	cghi: cgi headers that comes from cgi output,
*	and from mimetype of CGI handler */
Db *msgh;
/*	msgh: http header that should pass to fail or ok message. */

int authed = 0; /* authenticated. controled in hauth.c */
char *okip;

typedef struct laccess laccess;
struct laccess{
	char ip[16];
	long time;
};

/*	Note on cont
*	cont is used as a flag so that we can detect re-request using Keep-Alive
*	if (cont > 0)
*		re-request due to Kepp-Alive
*		so we should not reconfigure our namespace
*		"cont" means "continued"
*/

static	char *csquery(char*, char*, char*);
static	char *ndbquery(char *attr, char *value, char *attr);
static	void		becomenone(void);
static	int		nsconfig(HConnect *c, char *webbase);
static	void		dolisten(char*);
static	int		doreq(HConnect*);
static	int		send(HConnect*);
static 	void		doservice(char *ndir);
static	int		merge(char *mnt, char *root);
static	void	setcgienv(HConnect *c, Handle *hp);
static	int		ramfs(HConnect *c);

char *	addoptions(HConnect *c, char *uri);
static	int	creject(int ctl, char *ndir);
static	int	httpstatus(char *status);
static	int	hparse2env(HConnect *c, char *head);

char *usage=
	"usage: httpd [-Mmsuw] [-t timeout] [-p port] [-c certificate] [-n namespace] [-R rewrite]  [[httpd] dummy ..  incalldir]\n";

void
main(int argc, char **argv)
{
	char buf[256];
	char *incalldir = nil;
	fmtinstall('D', hdatefmt);
	fmtinstall('H', httpfmt);
	fmtinstall('U', hurlfmt);
	if(readconfig(httpdconf) < 0){
		syslog(0, HTTPLOG, "# error in reading config # %r");
		exits("readconfig");
	}
	//setstrconfig(&myname,"myname");
	setstrconfig(&webbase,"base");
	setstrconfig(&namespace,"namespace");
	setstrconfig(&rewrite,"rewrite");
	setstrconfig(&charset,"charset");

	setintconfig(&waittimeout1, "waittimeout1");
	setintconfig(&waittimeout2, "waittimeout2");
	setintconfig(&parsetimeout, "parsetimeout");
	setintconfig(&cgitimeout, "cgitimeout");
	setintconfig(&posttimeout, "posttimeout");
	setintconfig(&maxpost1, "maxpost1");
	setintconfig(&maxpost2, "maxpost2");
	maxpost1 *= 1024*1024;
	maxpost2 *= 1024*1024;
	setintconfig(&maxconnect, "maxconnect");
	setintconfig(&obstime, "obstime");
	setintconfig(&lockouttime,"lockouttime");
	setintconfig(&contmax,"contmax");

	waittimeout = waittimeout1;

	ARGBEGIN{
	case 'n':
		namespace = ARGF();
		break;
	case 'p':
		port = ARGF(); /* http */
		break;
	case 'c':
		certificate = ARGF();
		cert = readcert(certificate, &certlen);
		if(cert == nil)
			sysfatal("reading certificate: %r");
		break;
	case 't':
		cgitimeout = atoi(ARGF());
		break;
	case 'R':
		rewrite = ARGF();
		break;
	case 's':
		server = 1;
		break;
	case 'm':
		mntok = 1;
		break;
	case 'u':
		usermode = 1;
		break;
	case 'w':
		noweb = 1;
		break;
	case 'M':
		monmode = 1;
		break;
	default:
		sysfatal(usage);
		break;
	}ARGEND

	if(argc > 0)
		incalldir = argv[argc-1];
	if((server && incalldir) || (!server && !incalldir))
		sysfatal("# error: inconsistent use of -s and incalldir: %s", usage);
	if(!server && monmode)
		sysfatal("# error: inconsistent use of -M: %s", usage);

	if(webbase[0] != '/')
		sysfatal("webbase must begin with '/'");
	
	snprint(buf, sizeof buf, "%s/mnt", webbase);
	webmnt = estrdup(buf);

	if(namespace[0] != '/')
		sysfatal("namespace must begin with `/'");

	if(rewrite[0] != '/')
		sysfatal("rewrite must begin with `/'");


	scheme = cert?"https":"http";
	if(!server){
		if(usermode)
			sysfatal("inconsistent options");
		time(nil);
		contentinit();
		doservice(incalldir);
		exits(nil);
	}

	if(!monmode){
		switch(rfork(RFNOTEG|RFPROC|RFFDG|RFNAMEG)) {
		case -1:
			sysfatal("fork");
		case 0:
			break;
		default:
			exits(nil);
		}
	}

	/*
	 * open all files we might need before castrating namespace
	 */
	time(nil);
	syslog(0, HTTPLOG, nil);

	contentinit();

	runas = getuser();
	if(!usermode && (strcmp(runas, "none") != 0)){
		runas = "none";
		if(!certificate)
			becomenone();
	}
	dolisten(netmkaddr("*", "tcp", port));
	exits(nil);
}

void
hbind(HConnect *c, char *name, char *old, int flag)
{
	if(bind(name, old, flag) < 0){
		logit(c,"error: bind %s %s %d # %r", name,old,flag);
		hfail(c, HInternal);
		exits("bind error");
	}
}

static void
becomenone(void)
{
	int fd;

	fd = open("#c/user", OWRITE);
	if(fd < 0 || write(fd, "none", strlen("none")) < 0)
		sysfatal("can't become none");
	close(fd);
	if(newns("none", nil) < 0)
		sysfatal("can't build normal namespace");
}

static HConnect*
mkconnect(void)
{
	HConnect *c;

	c = ezalloc(sizeof(HConnect));
	c->hpos = c->header;
	c->hstop = c->header;
	return c;
}

static HSPriv*
mkhspriv(void)
{
	HSPriv *p;

	p = ezalloc(sizeof(HSPriv));
	return p;
}

static void
dolisten(char *address)
{
	char ndir[NETPATHLEN], dir[NETPATHLEN], *p;
	char *netdir, netdirb[256];
	char *ip;
	int m, ctl, nctl, data, spotchk;
	TLSconn conn;
	laccess *lbuf;
	int lbufsize = 2*maxconnect;
	int lnext = 0;

	spotchk = 0;
	if(cert)
		syslog(0, HTTPLOG, "httpd starting on %s using %s", address, certificate);
	else
		syslog(0, HTTPLOG, "httpd starting on %s", address);
	ctl = announce(address, dir);
	if(ctl < 0){
		syslog(0, HTTPLOG, "error: can't announce on %s: %r", address);
		return;
	}
	//syslog(0,HTTPLOG,"%s %s", address, dir);
	/* 	then
	*	address is "tcp!*!80"
	*	dir is " /net/tcp/23"
	*	term% netstat -n
	*	tcp  23   web        Listen       80         0          ::
	*/
	strcpy(netdirb, dir);
	p = nil;
	if(netdirb[0] == '/'){
		p = strchr(netdirb+1, '/');
		if(p != nil)
			*p = '\0';
	}
	if(p == nil)
		strcpy(netdirb, "/net");
	netdir = netdirb; // this is "/net"
	lbuf = calloc(lbufsize, sizeof(laccess));
	if(lbuf == nil)
		sysfatal("Internal Server Error: lbuf no memory");

	okip = share(OKSIZE);
	if(!okip)
		sysfatal("500 Internal Server Error: segattach error");
	*okip = 0;

	for(;;){
		/*
		 *  wait for a call (or an error)
		 */
		nctl = listen(dir, ndir);
		/*	example:
		*	dir="/net/tcp/23"	# listen using this directory
		*	ndir="/net/tcp/33"	# new connection
		*	dir is a listenning directory
		*	ndir is the assigned directory
		*	nctl is opened in listen() by open("/net/tcp/23/listen", ORDWR)
		*	and should be closed in signle connection
		*	listen() reads "/net/tcp/23/listen" to find ndir
		*	listen() returns open fd of "/net/tcp/23/listen"
		*	Ref: /sys/src/libc/9sys/announce.c		*/
		if(nctl < 0){
			syslog(0, HTTPLOG, "error: can't listen on %s: %r", address);
			syslog(0, HTTPLOG, "error+ ctls = %d", ctl);
			return;
		}

		strcpy(myip,netaddr(ndir,"local"));

		/*	Some protection against DoS attack that intends to make
		*	huge numbers of connected port.
		*	Note that persistent connection loop cannot be detected here.
		*	Be careful not to reject requests from proxy server.	*/
		ip = netaddr(ndir,"remote");
		//syslog(0,HTTPLOG,"ip=%s okchk=%d",ip,okchk(ip)); //@
		if(okchk(ip) == 0 && maxconnect > 0){
			int i;
			int n = 0;
			long t;

			if(black(ip)){
				creject(ctl, ndir);
				close(nctl);
				syslog(0,"blacklist", "%s %s # reject", port, ip);
				continue;
			}

			/*	time(nil) returns local clock in starting.
			*	however in few 10 seconds it may change to the time based on NTP.
			*	then time returned by time(nil) will jump discontinuously.
			*	so we apply labs(t - lbuf[i].time) to protect the anomaly that comes
			*	from this phenomenon	*/

			lbuf[lnext].time = t = time(nil);
			strncpy(lbuf[lnext].ip,ip,16);
			lnext++;
			lnext %= lbufsize;

			for(i = 0; i < lbufsize; i++){
				if(labs(t - lbuf[i].time) < obstime && strncmp(lbuf[i].ip, ip,16) == 0){
					n++;
					if(n > maxconnect){
						/*	This condition is system dependent,
						*	therefore required tuning.	*/
						creject(ctl, ndir);
						close(nctl);
						putblist(ip); // and throw that into blacklist
						syslog(0,"blacklist", "%s %s # maxconnect", port, ip);
						continue;
					}
				}
			}
		}

		/*
		 *  start a process for the service
		 */
		switch(rfork(RFFDG|RFPROC|RFNOWAIT|RFNAMEG)){
		case -1:
			close(nctl);
			continue;
		case 0:
			/*
			*	see if we know the service requested
			*
			*	What is the first argument of accept() should be ?
			*	man dial(2) says, it is nctl
			*	however /sys/src/cmd/ip/httpd/httpd.c says, ctl
			 */
			data = accept(ctl, ndir);
			if(data < 0){
				syslog(0, HTTPLOG, "error: can't open %s/data: %r", ndir);
				exits(nil);
			}
			if(cert){
				memset(&conn, 0, sizeof(conn));
				conn.cert = cert;
				conn.certlen = certlen;
				data = tlsServer(data, &conn);
				if(data < 0){// added by Kenar
					syslog(0, HTTPLOG, "error: tlsServer certificate=%s: %r", certificate);
					exits(nil);
				}
			}
			dup(data, 0);
			dup(data, 1);
			dup(data, 2);
			close(data);
			close(ctl);
			close(nctl);

			if(!usermode && certificate)
				becomenone();
			doservice(ndir);

			exits(nil);
		default:
			close(nctl);
			break;
		}

		if(++spotchk > 50){
			spotchk = 0;
			contentinit();
//			urlinit();
//			statsinit();
		}
	}
	free(lbuf);
}

static void
doservice(char *ndir)
{
	HSPriv *hpp;
	HConnect *c;
	NetConnInfo *nci;
	int ok;
	char *uripath, *p;

	nci = getnetconninfo(ndir, -1);
	c = mkconnect();
	hpp = mkhspriv();
	hpp->remotesys = nci->rsys;
	hpp->remoteserv = nci->rserv;
	port=nci->lserv; // this is real port
	c->private = hpp;

	hinit(&c->hin, 0, Hread);
	hinit(&c->hout, 1, Hwrite);

	/*
	 * original comment for timeout is here:
	 * 	serve requests until a magic request.
	 * 	later requests have to come quickly.
	 * 	only works for http/1.1 or later.
	 * I don't know the meaning. the code was:
	 *	for(timeout = 15*60*1000; ; timeout = 15*1000){
	 * I replaced the code by simple one.
	 * -Kenar-
	 */

	/*	It's safe to prevent redirection loop here	*/
	for(;;){
		/* waiting requests from the clients */
		if(hparsereq(c,waittimeout *1000) <= 0) // wait timeout (in unit of milisec)
			exits(nil);
		if(ntossed > 3){ // attack or redirection loop.
			/*	3 may be sufficient for allowed toss count	*/
			syslog(0, HTTPLOG, "warning: redirection loop ?");
			hfail(c, HInternal);
			exits(nil);
		} 
		/*
		*	hparsereq() get basic parameters in the request:
		*	c->req.meth		# method "GET", "POST" ...
		*	c->req.vermaj
		*	c->req.vermin
		*	c->req.urihost		# nil, the value will be assigned by hparseheaders()
		*	c->req.uripath			# uripath: path part in req. URI
		*	c->req.options		# ;params?query
		*	c->req.params		# params	got from req. URI
		*	c->req.query		# query	got from req. URI
		*
		*	scheme://urihost:port/uripath;params?query
		*	we should use real port no.
		*
		*	and also we have:
		*	c->req.user	nil
		*	c->req.path	nil
		*	these value will be assigned next.
		*	Note that there is no scheme info in the request.
		*	-Kenar-
		*/
		c->req.user = nil;
		uripath = c->req.uripath;
		c->req.path = uripath;
		if(uripath[0] == '/' && uripath[1] == '~'){
			p = strchr(uripath+2,'/');
			if(p)
				*p = 0;
			c->req.user = hstrdup(c, uripath+2);
			if(p)
				*p = '/';
			c->req.path = p;
		}
		/*	In request
		*		http://host/~alice/foo
		*	then
		*		req.uripath = "/~alice/foo"
		*		req.user	= "alice"
		*		req.path	= "/foo"
		*	In request
		*		http://host/foo
		*	then
		*		req.uripath = "/foo"
		*		req.user	= nil
		*		req.path 	= "/foo"
		*	In request
		*		http://host/~/foo
		*	then
		*		req.uripath = "/~/foo"
		*		req.user	= ""
		*		req.path 	= "/foo"
		*	that is, uripath includes user info.
		*/


		/*	cgih must be initialized in every request */
		if(cgih)
			db_free(cgih);
		cgih = nil;
		if(msgh)
			db_free(msgh);
		msgh = nil;

		ok = doreq(c);
		//logit(c,"DEBUG:doservice:doreq(): closeit=%d ok=%d",c->head.closeit,ok);
		if(c->head.closeit || ok < 0)
			exits(nil);
		hreqcleanup(c);
	}
}


static int
doreq(HConnect *c)
{
	char *uripath, *origuripath, *p, *q, *host, *user;
	char virtualhost[100];
	static char *host0 = nil;
	static char *user0 = nil;
	char *webroot = nil;
	char buf[256];
	int status;
	int n;
	char *dom;
	char *myname; // server name. example: ar in ar.aichi-u.ac.jp
	char *padom; // parent dom. example: aichi-u.ac.jp

	dom = ndbquery("ip", myip,"dom");
	if(!dom){
		logit(c,"# no dom");
		sysfatal("no dom");
	}
	p = hstrdup(c,dom);
	q = strchr(p,'.');
	if(!q){
		logit(c,"# no '.' in dom");
		sysfatal("dom");
	}
    *q = 0;
	myname = p;
	padom = ++q;

	uripath = c->req.uripath;
	/*
	 * normal case is just file transfer
	 */
	if(cont > 0){
		logit(c, "continue: %d",cont);
		if(cont > contmax){ /* may be required for safty */
			logit(c, "continuation exceed max: %d uripath=%s",cont, uripath);
			return hfail(c, HInternal);
		}
		cont++;
	}

	/*
	*	Kenar complains:
	*	if(hparseheaders(c, 15*60*1000) < 0)
	*	why so long time for parsing headers?
	*	I replaced by:
	*/
	if(hparseheaders(c, parsetimeout*1000) < 0)
		exits("failed");
	/*
	*	for old client c->head.host may be nil
	*	we must supply the host name from ip value
	*/
	//logit(c,"DEBUG: head.host=%s", c->head.host);
	if(c->head.host == nil){
		p = ndbquery("ip", myip,"dom");
		if(p == nil){
			fprint(2, "ndbquery() returned nil\n");
			exits("ndbquery");
		}
		c->head.host = hstrdup(c, p);
	}
	if(c->head.host == nil){
		logit(c, "host: nil");
		exits("failed");
	}
	//logit(c,"head.host=%s", c->head.host);
	/*
	*	now we will have:
	*	c->head.host
	*		the value is already lowered.
	*	Let ar be a host name and aichi-u.ac.jp is the dom
	*	then head.host is something like:
	*	ar
	*	ar.aichi-u.ac.jp
	*	ar.aichi-u.ac.jp:80	# some browsers send HTTP Host header with port number
	*
	*	c->head.host gives hhost (header host) info for uriroot
	*	we can establish uriroot
	*/
	host = c->head.host;
	n = strlen(scheme) + strlen(host) + 10;
	uriroot = halloc(c, n);
	/*	in this stage port part exist in host	*/
	snprint(uriroot, n, "%s://%s", scheme, host);

	/*
	*	Let's remove ":port" here from urihost
	*	although this should be done in library
	*	-Kenar-
	*/
	host = hstrdup(c, host);
	p = strchr(host,':');
	if(p)
		*p = 0;

	if(!http11(c) && !c->head.persist)
		c->head.closeit = 1;
	logit(c, "%s %s HTTP/%d.%d", c->req.meth, c->req.ouri, c->req.vermaj, c->req.vermin);
	//logit(c, "DEBUG: %s", c->header);
	/*
	*	we can remove padom part from host
	*	so that we can simplify management of /sys/lib/httpd.rewrite
	*	now we don't need full name for virtual host in this file.
	*	be careful the host name might be IP.
	*/
	if(strlen(host)>strlen(padom)){
		char *p, buf[256];
		snprint(buf, sizeof buf, ".%s", padom);
		p = host + strlen(host) - strlen(buf);
		if(strcmp(p, buf) == 0)
			*p = 0;
	}
	/*	then host is "ar in ar.aichi-u.ac.jp" for example
	*	on the other hand, host "202.250.160.40" remains as it is	*/

	/*
	*	We must examine that the connection is really continued ?
	*	How to determine the connection should enter new name space?
	*	How to enter new name space?
	*	HTTP referer defined in HTTP/1.0 and HTTP/1.1 might be helpful.
	*	referer is http://host/path if URI is from href in html document
	*	example:	http://host/~alice/photos/
	*	According to RFC, referer can be set so that it is not sent for privacy
	*	But referer is nil if URI is from keyboard input,
	*	therefore referer is not acceptable for this purpose.
	*	
	*	We claim to resend if (host, user) pair has changed
	*	and close the connection
	*
	*	schemes and ports concern nothing to this problem.
	*	because different servers care for them
	*
	*	-Kenar-
	*/


	/*
	*	Problem with favicon.ico
	*
	*	Some browsers[1] always tries to get /favicon.ico
	*	in accessing directories,
	*	even if in accessing user space such as "/~alice/foo/bar/"
	*	and even if the absence of "/favicon.ico" is already have 
	*	been informed by 404 error message.
	*	This is annoying, and farther more, makes a problem in case
	*	that we should keep connection with the browser.
	*	Note that Pegasus 2.1 disconnects in constructing new name space.
	*	Pegasus 2.2 simply tries to send a file "/~alice/favicon.ico" instead of 
	*	"/favicon.ico" and keeps connection in case that Pegasus is in
	*	alice's namespace.
	*
	*	I hope the day comes such a nasty trick is wiped out.
	*
	*	[1]: Safari for  example
	*
	*	-Kenar-
	*/
	user = c->req.user;
	//logit(c,"DEBUG: doreq: user=%s user0=%s uri=%s",user,user0,c->req.uripath);
	if(user0 && user== nil){
		/* 	the Pegasus is alreay in user0 space, but the request is
		*	to the host owner's space */
		int m;
		char *p;
		m = strlen(c->req.uripath) + 10;
		p = halloc(c,m);
		snprint(p,m,"/doc/~%s",c->req.uripath);
		/*	if req.uripath is "/favicon.ico", then p is
		*	"/doc/~/favicon.ico"		*/
		if(executable(p)){
			/* don't send executable file of real host */
			hfail(c, HNotFound, p);
			return 0;
		}
		if(strstr("GET,HEAD", c->req.meth) == 0){
			logit(c, "unallowed: %s %s", c->req.meth, p);
			return hunallowed(c, "GET, HEAD");
		}
		if(access(p,0) != -1){
			sendit(c,p, HCNo);
			c->head.closeit = 0;
			logit(c,"send %s and reconnect",p);
			return 0; //continue
		}
	}

	if(cont){
		if(xstrcmp(host, host0) != 0 || xstrcmp(user, user0) != 0){
			/*	claim to resend */
			//logit(c,"DEBUG: host:%s -> %s user:%s -> %s", host0,host,user0,user);
			logit(c,"resend to:  %s%s", uriroot, c->req.ouri);
			if(strcmp(c->req.meth,"POST")==0){
				c->head.closeit = 0;
				hredir(c, "303 See Other", c->req.uripath, c->req.options);
				return 0; //continue
			}
			c->head.closeit = 1;
			hmoved(c, c->req.uripath, c->req.options);
			return -1; //exit
		}
	}
	else{
		host0 = estrdup(host);
		if(user) user0 = estrdup(user);
		/*	don't use hstrdup because memory is initialized
		*	in continued connection		*/
	}

	if((strcmp(scheme,"https") == 0 && strcmp(port,httpsport) == 0)
	|| (strcmp(scheme,"http") == 0 && strcmp(port,httpport) == 0))
		snprint(virtualhost, sizeof virtualhost, "%s://%s/", scheme, host);
	else
		snprint(virtualhost, sizeof virtualhost, "%s://%s:%s/", scheme, host, port);
	origuripath = c->req.uripath;
	uripath = nil;
	if(user == nil)
		uripath = redirect(c, rewrite, virtualhost, &webroot, scheme, &hostroot);
	#ifdef NOTE
	*	we have three cases of uri
	*	1. 	scheme://host/foo
	*	2. 	/foo
	*	3. 	nil
	#endif
	if(uripath){
		/*	Incoming request is limited by /sys/lib/httpd.rewrite to our host	*/
		/*
		if(uripath[0] == '*'){
			webroot = hstrdup(c, uripath + 1);
			binding = 1;
		}
		*/
		if(webroot)
			binding = 1;
	}
	else{
		/*
		*	All request will come here except to the host listed in /sys/lib/httpd.rewrite.
		*
		*	we can reject faked accesses here;
		*	if uripath is nil and host is not myname, we may reject.
		*	example:
		*	ar	# sysname
		*	plan9	# virtual host of ar. Request to plan9 will not come here.
		*	aichi-u.ac.jp	# padom
		*	now ar.aichi-u.ac.jp and plan9.aichi-u.ac.jp are reduced to
		*	ar and plan9 respectively.
		*	On the other hand
		*	www.faked.com
		*	is kept untached.
		*	The condition to reject is:
		*	(strcmp(host, myname) == 0 || strcmp(host, myip) == 0) != 0
		*/
		if(strcmp(host, myname) != 0 && strcmp(host, myip) != 0){
			if(user)
				logit(c, "denied: %s/~%s", host, user);
			else
				logit(c, "denied: [%s] [%s]", host, myname);
			return hfail(c, HNotFound, host);
		}
		/*	redirection is examined only for real host	*/
		uripath = redirect(c, rewrite, origuripath, &webroot, scheme, &hostroot);
		//logit(c,"uri=%s origuri=%s webroot=%s", uri, origuri, webroot);
		if(uripath && webroot){
			/*	we just accept the incoming request	*/
			c->req.path=uripath;
			if(c->req.user)
				binding = 2;
			else
				binding = 1;
		}

	}

	mnt = 0;
	if(webroot){
		#ifdef NOTE
			This case is '*' redirect in /sys/lib/httpd.rewrite such as:
			http://pegasus  */usr/someone/www
			/~alice	*/usr/alice/www
			/car		*/usr/carol/www
			/		*/usr/bob/www
		#endif
		if(!cont){
			hbind(c, webroot, webmnt , MREPL);
			mnt = 1;
			hroot = webroot;
		}
	}
	else if(uripath != nil){ /* Let the client know the new address */
		char *o;
		o = c->req.options;
		logit(c,"redirect1 to:  %s%s ntossed=%d", uripath, o, ntossed);
		ntossed++;
		return hmoved(c, uripath, o);
	}


	/*
	* 	Here we support such as http://ar/~alice/abc
	*	where "ar" must be real host and need noweb == 0.
	*	Then we bind /usr/alice/web to /usr/web/mnt
	*	and change request document to /abc
	*	This redirection is automatically done without rewrite-file
	*
	*	-Kenar-
	*/
	uripath = c->req.uripath;
	if(noweb == 0 && c->req.user && binding != 1){
		/*	then uri[0]=='/' && uri[1]=='~'	*/
		char *t;
		/* 	if uripath is "/~alice" for example,
			let the client resend the correct uri "/~alice/"
			-Kenar-	*/
		t = strchr(uripath+2, '/');
		if(!t){
			char *o;
			o = c->req.options;
			snprint(buf, sizeof buf, "%s/", uripath);
			logit(c, "redirect2 to: %s%s", buf, o);
			return hmoved(c, buf, o);
		}

		if(!cont && !mnt){
			/*
			*	Pegasus 2.2 support global space such as
			*	bind $hostroot/doc/~ $web/doc/~
			*	in rc notation
			*	where
			*		hostroot=/usr/bob/www
			*		web=/usr/alice/web
			*	for example
			*
			*	-Kenar-
			*/

			char *user;
			char *path,*hpath;
			int n;

			n = strlen(hostroot) + 10;
			hpath = halloc(c, n);
			snprint(hpath, n, "%s/doc/~", hostroot);

			user = c->req.user;
			if(user && *user){
				n = strlen("/usr//web") + strlen(user) + 10;
				path = halloc(c, n);
				snprint(path, n, "/usr/%s/web/doc/~", user);
				bind(hpath, path, MREPL); // error OK, don't mind

				/*	bind $web $webmnt	*/
				snprint(path, n, "/usr/%s/web", user);
				hbind(c, path, webmnt, MREPL);
				mnt = 1;
				hroot = path;
			}
		}

		/* 	then if c->req.uripath is "/~alice/abc", the path is "/abc" 
			for the request http://ar/~alice/abc
			We need not wary about "../" in URI such as "/~alice/../",
			it reduces to "/" as it should be.
			-Kenar-	*/
	}

	/* namespace configuration; hide everything out of webbase */
	if(!cont){
		if(addns(runas, namespace) < 0){
			logit(c, "error: addns %s # %r", namespace);
			sysfatal("can't build httpd namespace");
		}
		if(mnt){
			char nsbuf[32];
			/* 	strlen("namespace_65536") is 15	*/
			nsconfig(c, webbase);
			snprint(nsbuf, sizeof nsbuf, "/etc/namespace_%s", port);
			//logit(c,"etc/namespace: %s", nsbuf);
			if(mntok)
				addns(runas, nsbuf);
			else
				chgns(nsbuf);
		}
		cont = 1;
	}
	chdir("/");

	/*	Namespace configuration is completed.
	*	Now we are in the configured namespace.	*/

	//nsdump("/etc");

	return send(c);
}

/*	# return values
*	0,+1: OK
*	-1: NG
*	the return value should be 0 or -1
*	inconsitently used OK value. original confusion is in sentdit()
*	-Kenar-	*/
static int
send(HConnect *c)
{
	Dir *dir;
	Handle *hp;
	HAuth acheck;
	char *p;
	char *intpath; // internal path (path from httpd root)
	char *uripath, *newuri;
	char *path; // path from document root
	int n, ok;
	ulong dirmode;
	//char buf[256];
	int status;
	HSPriv *hpp = c->private;

	/*	Namespace configuration is completed.
	*	Now we are in the configured namespace.	*/

	/*	DEBUG: monitor header */
	if(c->head.expectother || c->head.expectcont)
		//logit(c, "DEBUG: send: expect* HEAD=%s", c->header);


	/* 	RFC2068

	8.2 Message Transmission Requirements

	- An HTTP/1.1 (or later) server that receives a request from a
	   HTTP/1.0 (or earlier) client MUST NOT transmit the 100 (continue)
	   response; it SHOULD either wait for the request to be completed
	   normally (thus avoiding an interrupted request) or close the
	   connection prematurely.
	-   Upon receiving a method subject to these requirements from an
	   HTTP/1.1 (or later) client, an HTTP/1.1 (or later) server MUST either
	   respond with 100 (Continue) status and continue to read from the
	   input stream, or respond with an error status. If it responds with an
	   error status, it MAY close the transport (TCP) connection or it MAY
	   continue to read and discard the rest of the request. It MUST NOT
	   perform the requested method if it returns an error status.

	10.1.1 100 Continue

	 The client may continue with its request. This interim response is
	 used to inform the client that the initial part of the request has
	 been received and has not yet been rejected by the server. The client
	 SHOULD continue by sending the remainder of the request or, if the
	 request has already been completed, ignore this response. The server
	 MUST send a final response after the request has been completed.

	*/
	if(c->head.expectcont && http11(c))
		hokmsg(c, "100 Continue");

	if(c->head.expectother)
		return hfail(c, HExpectFail);

	path = c->req.path;
	uripath = c->req.uripath;
	/*	"path" is a path from document root
	*	"uripath" is a uri oriented path, a path part in http://host/path
	*	The difference apeares in regular user, these values are:
	*	path=/foo, uripath=/~alice/foo	*/
	//logit(c, "DEBUG: send: path=%s, uripath=%s", path, uripath);

	newuri = redirect(c, "/etc/rewrite", path, nil, scheme, &hostroot);
	#ifdef NOTE
	*	we have three cases of newuri
	*	1. 	scheme://host/path
	*	2. 	/foo or /~alice/foo
	*	3. nil
	*	examples:
	*	uripath=="/~alice" and newuri=="/~alice"
	#endif

	//logit(c,"DEBUG: send: path=%s newuri=%s", path, newuri);

	if(newuri){
			char *o;
			o = c->req.options;
			logit(c, "moved to: %s%s", newuri, o);
			ntossed++;
			return hmoved(c, newuri, o);
	}			

	/*
	 * check for directory/file mismatch with trailing /,
	 * and send any redirections.
	 */
	n = strlen("/doc") + strlen(path) + strlen("/index.html") + 10;
	/* make a room to insert index.html */
	intpath = halloc(c, n);
	strcpy(intpath, "/doc");
	strcat(intpath,path);


	/*
	*	In case of GET and HEAD method,
	*	we examine if it is a directory.
	*	if directory we must required the client to resend the request
	*	with a '/' at the end of path name.
	*	
	*	note that the path need not exist
	*	because it may be handled by handler
	*/
	if((dir = dirstat(intpath)) != nil){
		dirmode = dir->mode;
		free(dir);
		if(dirmode & DMDIR){
			/*
			*	if path does not end with '/'
			*	then let the client resent the uri with "/"
			*	Don't replace by the code such as the following
			*	because HTTP/1.0 client will not work
			*	They hate redirection
			*	-Kenar-
			*
			* 	if(path[strlen(path)-1] == '/')
			*		snprint(buf, sizeof buf, "%sindex.html", uri);
			* 	else
			*		snprint(buf, sizeof buf, "%s/index.html", uri);
			* 	return hmoved(c, buf);
			*
			*	Problem with redirection:
			*	Some browsers[1] disconnects in redirection
			*	even if the redirection is
			*	/foo => /foo/
			*	and even if "Connection: Keep-Alive" is sent to the brower
			*	Therefore be carefull in writing HTML code.
			*
			*	[1]: Safari for example
			*	Firefox does not have this problem.
			*/
			if(uripath[strlen(uripath)-1] != '/'){
				int n;
				char *b;
				n = strlen(uripath) + 5;
				b = malloc(n);
				snprint(b, n, "%s/", uripath);
				logit(c, "redirect3 to: %s", b);
				//c->head.closeit = 0;
				c->head.host = nil;
				status = hmoved(c, b, c->req.options);
				if(status)
					logit(c, "#ERROR: redirect3: hmoved");
				/* 	internal error if status is -1
				*	then the connection should be freed
				*	-Kenar-	*/
				free(b);
				return status;
			}
			/*
			*	here the form of uripath is "/foo/" if "foo" is a directory
			*	that is, uripath ends with '/'
			*	-Kenar-
			*/


		}else if((p = strchr(intpath, '\0')) > intpath && p[-1] == '/'){
			/*
			*	This condition means:
			*	path is a file and ends with '/' 
			*	for example,
			*		http://ar.aichi-u.ac.jp/index.html/
			*	Then let the client to resend
			*		http://ar.aichi-u.ac.jp/index.html
			*	-Kenar-
			*/
			*strrchr(uripath, '/') = '\0';
			logit(c, "%s hmoved", uripath);
			return hmoved(c, uripath, c->req.options);

			/*	The redirection might be harmful in CGI/1.1 support.
			*	Because CGI/1.1 supports uri such as
			*		http://host/cgi-bin/foo.cgi/bar
			*	where "foo.cgi" is a file. The special case is
			*		http://host/cgi-bin/foo.cgi/
			*/
		}
	}

	if(uripath[strlen(uripath)-1] == '/')
		strcat(intpath, "index.html");
	c->req.intpath = intpath;

	/*	Now we have gotten normalized URI. The results are in uripath and intpath
	*	For the request
	*		http://host/~alice/foo
	*	if foo is a file
	*		path = "/foo"
	*		intpath = "/doc/foo"
	*		uripath = "/~alice/foo"
	*	on the other hand, if foo is a direcory, it is redirected to
	*		http://host/~alice/foo/
	*	and
	*		path = "/foo/"
	*		intpath = "/doc/foo/index.html"
	*		uripath = "/~alice/foo/"
	*/



	/*
	*	access control must be the last
	*	because it is delicate to handle last slash
	*	so full path is required to evaluate
	*	-Kenar-	*/

	if(ipchk(c, "/etc/allow", hpp->remotesys, path) < 0){
		logit(c, "ipchk: unallowed %s %s", hpp->remotesys, path);
		return hfail(c, HForbidden, path);
	}

#ifdef WIN2000DAV
	/*	Web servers that supports Win2000 client must allow
	*	OPTIONS without authentication with
	*	"MS-Author-Via: DAV"
	*	header. I dislike to bring this feature to Pegasus specification.
	*	Therefore Win2000 support went to a compiler option.
	*	-Kenar-	*/
	if(strcmp(c->req.meth,"OPTIONS") == 0){
		db_put(&msgh, "DAV:", "1, 2");
		db_put(&msgh, "MS-Author-Via:", "DAV");
		db_put(&msgh, "Allow:",
		"OPTIONS, GET, HEAD, PUT, DELETE, PROPFIND, PROPPATCH, MKCOL, COPY, MOVE, LOCK, UNLOCK");
		return hokmsg(c, "200 OK");
	}
#endif


	ok = usrchk(c, &acheck, "/etc/passwd", path);
	if(ok <= 0){
		logit(c, "unallowed: %s", path);
		/* the return value should be -1 to make clean connection */
		return -1;
	}
	/* as the side effect, we have authed==1 if autherized */
	if(authed)
		updateOkip(hpp->remotesys);

	/*	Now every redirection is finished and we have ready.
	*	requested path is already converted to absolute path.
	*	values to be used are:
	*	path: path to the file to send. where path is in httpd name space
	*	fd: open file descriptor of path
	*
	* 	CGI is supported using exechandler.
	*	The handler configuration file is /etc/handler in service space.
	*	Handler will be obtained from requested path.
	*	Handler must be exec permitted.
	*	-Kenar-
	*/

	ntossed = 0;

	//logit(c,"DEBUG: intpath=%s", intpath);

	/*	call exechandler	*/
	hp = handler(c, "/etc/handler", intpath + 4);
	if(hp){
		char **argv;

		argv = hp->argv;
		if(strcmp(hp->mimetype,"-") != 0)
			db_put(&cgih, "Content-Type:",hp->mimetype);

		// logit(c,"DEBUG: argv[0]=%s %s",argv[0], argv[1]);

		if(executable(argv[0])){
			int status;
			char *p;

			logit(c,"exec %s", argv[0]);
			if(c->head.contlen != 0 || (c->head.transenc && (
				cistrcmp(c->head.transenc->s,"chunked") == 0))){
				int maxpost;
				maxpost = authed?maxpost2:maxpost1;
				//logit(c,"receiving %d bytes", c->head.contlen);
				alarm(posttimeout*1000);
				status = acceptpost(c, maxpost);
				alarm(0);
				if(status  < 0)
					return -1; // make clean reconnection
			}

			if(hp->hctl == HCWild)
				hp->hctl = HCPlus;
			else{
				/* we disallow files starting with `.' */
				p = strrchr(intpath,'/');
				if(p[1] == '.'){
					logit(c, "%s forbidden", path);
					return hfail(c, HForbidden, path);
				}

				if(strcmp(c->req.meth,"OPTIONS") == 0){
					db_put(&msgh, "Allow:", "GET, HEAD, POST, OPTIONS");
					logit(c, "OPTIONS by unauth client");
					return hokmsg(c, "200 OK");
				}
				if(strstr("GET,HEAD,POST", c->req.meth) == 0){
					logit(c, "unallowed: %s %s", c->req.meth, p);
					return hunallowed(c, "GET, HEAD");
				}
			}

			/* chdir to the target dir, taget is intpath in httpd space	*/
			p = strrchr(intpath,'/');
			*p = 0;
			if(chdir(intpath) < 0)
				chdir("/doc");
			*p = '/';
			setcgienv(c, hp);
			ramfs(c);

			status = execute(c,argv, hp->hctl);
			unmount(nil, "/tmp");
			return status;
		}
	}

	/*	We might be prompted to prohibit to send the content
	*	of executable file for security, and write following codes:
	*	if(executable(intpath)){
	*		hfail(c, HForbidden);
	*		logit(c,"%s is executable, not sent", intpath);
	*		return 0; // continue
	*	}
	*	But then, Pegasus rejects sending some files in netlib.
	*	Therefore I disabled the codes above.
	*	-Kenar-	*/

	/* we disallow files starting with `.' */
	p = strrchr(intpath,'/');
	if(p[1] == '.'){
		logit(c, "%s forbidden", path);
		return hfail(c, HForbidden, path);
	}

	if(strcmp(c->req.meth,"OPTIONS") == 0){
		db_put(&msgh, "Allow:", "GET, HEAD, POST, OPTIONS");
		logit(c, "OPTIONS by unauth client");
		return hokmsg(c, "200 OK");
	}

	if(strstr("GET,HEAD", c->req.meth) == 0){
		logit(c, "unallowed: %s %s", c->req.meth, p);
		return hunallowed(c, "GET, HEAD");
	}

	//logit(c,"sending %s", intpath);
	status = sendit(c, intpath, HCNo);
	return status; // +1: success;  -1: fail
}

/* ------- Kenar's misc -------- */


/*	we are in emergency.
*	therefore we send a simple message to the client. sorry	*/
static int
creject(int ctl, char *ndir)
{
	int data;
	int status = 0;
	char *msg ="HTTP/1.1 503 Service Unavailable\r\n"
				"Server: Plan9\r\n"
				"Content-Type: text/plain\r\n"
				"Content-Length: 5\r\n"
				"Connection: close\r\n"
				"\r\n"
				"Busy\n";

	switch(rfork(RFFDG|RFPROC|RFNOWAIT|RFNAMEG)){
	case -1:
		status = -1;
		break;
	case 0:
		/*
		 *  see if we know the service requested
		*
		*	What is the first argument of accept should be ?
		*	man dial(2) says, it is nctl
		*	however /sys/src/cmd/ip/httpd/httpd.c says, ctl
		 */
		data = accept(ctl, ndir);
		if(data < 0){
			syslog(0, HTTPLOG, "error: can't open %s/data: %r", ndir);
			exits(nil);
		}
		write(data, msg, strlen(msg));
		close(data);
		exits(nil);
	default:
		break;
	}
	return status;
}

static char*
ndbquery(char *attr, char *value, char *rattr)
{
	char *p;
	Ndb *db;
	Ndbs s;
	Ndbtuple *t;

	db = ndbopen("/lib/ndb/local");
	if(db == 0){
		fprint(2, "no db files\n");
		exits("no db");
	}
	p = ndbgetvalue(db, &s, attr, value, rattr, &t);
	ndbclose(db);
	if(p){
		ndbfree(t);
		return p;
	}
	return nil;
}

/*
*	configure service namespace
*	don't put error handling because "/mnt" represents here users webroot
*	such as /usr/alice/web/
*	she might not have "bin" or other things.
*/
static int
nsconfig(HConnect *c, char *webbase)
{
	char *wmnt, *wetc, *wbin, *wmetc, *wmbinrc, *wmbinobj;

	wmnt = hcat2(c, webbase,"/mnt");
	hbind(c, wmnt, webbase, MAFTER);
	wetc = hcat2(c, webbase, "/etc");
	wbin = hcat2(c, webbase,"/bin");
	wmetc = hcat2(c, webbase, "/mnt/etc");
	wmbinrc = hcat2(c, webbase, "/mnt/bin/rc");
	wmbinobj = hcat3(c, webbase, "/mnt/bin/",getenv("objtype"));
	/*	 the unmount below is the guard when next three binds fail.
	*	for example, /usr/web/mnt/etc may be absent	*/
	bind(wmetc, wetc, MAFTER); // error OK, don't mind
	bind(wmbinrc, wbin, MAFTER); // error OK, don't mind
	bind(wmbinobj, wbin, MAFTER); // error OK, don't mind
	unmount(nil, wmnt);
	hbind(c, webbase, "/", MREPL);
	return 0;
}


static void
setcgienv(HConnect *c, Handle *hp)
{
	HSPriv *hpp = c->private;
	char buf[256];
	char *cputype, *name, *pathinfo, *user, *script;
	char *timezone;
	int n, ch;
	char *p;

	//logit(c,"DEBUG: setcgienv");

	cputype = getenv("cputype");
	timezone = getenv("timezone");

	rfork(RFCENVG);

	/*	Putting header info into environment
	*	must be the first. (for security)
	*	-Kenar-	*/
	ch = *(c->hstop);
	*(c->hstop) = 0;
	setenv("HTTP_HEADER", (char*)c->header);
	hparse2env(c, (char*)c->header);
	*(c->hstop) = ch;

	setenv("HTTP_SCHEME", scheme); // Pegasus
	setenv("DOCUMENT_ROOT", "/doc"); // CGI/1.1
	setenv("GATEWAY_INTERFACE", "CGI/1.1"); // CGI/1.1
	pathinfo = c->req.uripath;
	setenv("PATH_INFO", pathinfo); // CGI/1.1


	/*	PATH_INFO, PATH_TRANSLATED and SCRIPT_NAME for CGI/1.1
	*	Look comments in handler.c	*/
	user = c->req.user;
	script = hp->scriptname;
	pathinfo = hp->pathinfo;
	/*	We have four cases in combination with
	*		user == nil or not
	*		scriptname == nil or not	*/
	if(!pathinfo)
		pathinfo = "";
	if(*pathinfo){
		n = strlen(pathinfo) + 10;
		p = halloc(c,n);
		snprint(p, n, "/doc%s", pathinfo);
		setenv("PATH_TRANSLATED", p); // CGI/1.1
	}
	else
		setenv("PATH_TRANSLATED", ""); // CGI/1.1

	if(script)
		setenv("PATH_INFO", pathinfo); // CGI/1.1
	else if(user){
		n = strlen(user) + strlen(pathinfo) + 10;
		p = halloc(c,n);
		snprint(p, n, "/~%s%s", user, pathinfo);
		setenv("PATH_INFO", p);
	}
	else
		setenv("PATH_INFO", pathinfo);

	if(script){ // CGI/1.1
		if(user){
			n = strlen(user) + strlen(script) + 10;
			p = halloc(c,n);
			snprint(p, n, "/~%s%s", user, script);
			setenv("SCRIPT_NAME", p);
		}
		else
			setenv("SCRIPT_NAME", script);
	}
	else
		setenv("SCRIPT_NAME", "");

	setenv("REMOTE_ADDR", hpp->remotesys); // CGI/1.1
	setenv("REMOTE_HOST", hpp->remotesys); // CGI/1.1
	setenv("REQUEST_PATH", c->req.path);


	setenv("REQUEST_METHOD",c->req.meth); // CGI/1.1
	
	snprint(buf, sizeof buf, "%d.%d", c->req.vermaj,c->req.vermin);
	setenv("REQUEST_VERSION", buf);
	setenv("REQUEST_URI",c->req.ouri); // compatible with Apache
	setenv("AUTH_TYPE", c->head.authtype); // CGI/1.1
	setenv("REMOTE_USER", c->head.authuser); // CGI/1.1
	setenv("AUTH_PASS", c->head.authpass);
	snprint(buf, sizeof buf, "%uld", c->head.contlen);
	setenv("CONTENT_LENGTH", buf); // CGI/1.1
	setenv("CONTENT_TYPE", c->head.conttype); // CGI/1.1
	setenv("REQUEST_USER", c->req.user);
	setenv("SERVER_DOMAIN", ndbquery("ip",myip,"dom"));
	strncpy(buf,c->head.host,sizeof(buf));
	p = strchr(buf,':');
	if(p)
		*p = 0;
	setenv("SERVER_NAME", buf); // CGI/1.1

	setenv("SERVER_PORT", port); // CGI/1.1
	setenv("SERVER_PROTOCOL",hversion); // CGI/1.1
	setenv("SERVER_SOFTWARE", serversoft); // CGI/1.1
	setenv("QUERY_STRING", c->req.query); // CGI/1.1

	qs2env(c, c->req.query); // Pegasus Special
	setenv("request", c->req.path);
	setenv("target", (c->req.intpath) + 4);
	name = strrchr(c->req.path, '/');
	setenv("name", name+1);
	setenv("query", c->req.query);
	snprint(buf, sizeof buf, "%D", time(nil));
	setenv("DATE", buf);
	setenv("home", "/doc");
	setenv("cputype", cputype);
	setenv("objtype", cputype);
	snprint(buf, sizeof(buf), "%d", getpid());
	setenv("hpid", buf);
	setenv("hroot", hroot); // httpd root
	setenv("timezone", timezone);
}

static int
ramfs(HConnect *c)
{
	int pid, status;
	char msg[ERRMAX];
	/*	ramfs service to the child	*/
	switch(pid = rfork(RFFDG|RFPROC)){ /* assign = */
	case -1:
		logit(c, "error: fork # %r");
		hfail(c, HInternal);
		return -1;
	case 0:
		close(0);
		close(1);
		close(2);// prevent "ramfs: unmount read" message
		execl("/bin/ramfs","ramfs","-p",0);
		/* not reached */
		logit(c, "error: unable to execl /bin/ramfs %r");
		exits(nil);
	default:
		break;
	}
	
	/* ensure "/bin/ramfs" finished */
	status = waitfor(pid, msg);
	if(status < 0){
		logit(c,"error: /bin/ramfs: no children to wait %d", pid);
		return -1;
	}
	return 0;
}



/*	mime parse is not easy.
*	and probably meaningless to put the info int env.
*	so I chose a simple way
*	I parsed only
*		key:val
*	pairs, and
*		putenv(key, val)
*	-Kenar-	*/
static int
hparse2env(HConnect *c, char *head)
{
	char *p,*s;
	char **buf;
	char *kv[2];
	char buf1[64];
	int k,n,m;

	p = hstrdup(c, head);
	/* skip a line */
	p = strchr(p, '\n');
	if(!p)
		return -1;
	p++;
	n = 0;
	for(s = p; *s; s++)
		if(*s == '\n') n++;
	buf = halloc(c, (n + 1) * sizeof(char*));
	n = getfields(p, buf, n + 1, 0, "\n");
	
	/* strip the trailing CR */
	for(k = 0; k < n; k++){
		p = strchr(buf[k], '\r');
		if(p)
			*p = 0;
	}

	/*	get
	*		key: val
	*	and trim, then putenv	*/
	for(k = 0; k < n; k++){
		m = getfields(buf[k], kv, 2, 0, ":");
		if(m != 2)
			continue;
		for(p = kv[0]; *p; p++)
			*p = toupper(*p);
		for(p = kv[0]; *p; p++)
			if(*p == '-') *p = '_';
		snprint(buf1, sizeof buf1, "HTTP_%s", trim(kv[0]));
		setenv(buf1, trim(kv[1]));
	}
	return 0;
}

