#pragma	lib	"libhttpd.a"
#pragma	src	"libhttpd"


/*	basic constant	*/
extern char *serversoft;
extern char *HTTPLOG;
extern char *uriroot;

#define httpport "80"
#define httpsport "443"

typedef struct HConnect		HConnect;
typedef struct HContent		HContent;
typedef struct HContents	HContents;
typedef struct HETag		HETag;
typedef struct HFields		HFields;
typedef struct Hio		Hio;
typedef struct Htmlesc		Htmlesc;
typedef struct HttpHead		HttpHead;
typedef struct HttpReq		HttpReq;
typedef struct HRange		HRange;
typedef struct HSPairs		HSPairs;
typedef struct Bin		Bin;
typedef struct Handle Handle;

enum
{
	HMaxWord	= 32*1024,
	HBufSize	= 32*1024,

	/*
	 * error messages
	 */
	HInternal	= 0,
	HCGIErr,
	HTempFail,
	HUnimp,
	HCgiTimeout,
	HBadReq,
	HBadSearch,
	HNotFound,
	HSyntax,
	HNoSearch,
	HUnauth,
	HNoData,
	HExpectFail,
	HForbidden,
	HUnkVers,
	HBadCont,
	HNotAccept,
	HPrecond,
	HReqSize,
	HReqRange,
	HUnavailable,
	HOK,
};

/*
 * table of html character escape codes
 */
struct Htmlesc
{
	char		*name;
	Rune		value;
};

struct HContent
{
	HContent	*next;
	char		*generic;
	char		*specific;
	float		q;			/* desirability of this kind of file */
	int		mxb;			/* max uchars until worthless */
};

struct HContents
{
	HContent	*type;
	HContent	 *encoding;
};

/*
 * generic http header with a list of tokens,
 * each with an optional list of parameters
 */
struct HFields
{
	char		*s;
	HSPairs		*params;
	HFields		*next;
};

/*
 * list of pairs a strings
 * used for tag=val pairs for a query or form submission,
 * and attribute=value pairs in headers.
 */
struct HSPairs
{
	char		*s;
	char		*t;
	HSPairs		*next;
};

/*
 * byte ranges within a file
 */
struct HRange
{
	int		suffix;			/* is this a suffix request? */
	ulong		start;
	ulong		stop;			/* ~0UL -> not given */
	HRange		*next;
};

/*
 * list of http/1.1 entity tags
 */
struct HETag
{
	char		*etag;
	int		weak;
	HETag		*next;
};

/*
 * HTTP custom IO
 * supports chunked transfer encoding
 * and initialization of the input buffer from a string.
 */
enum
{
	Hnone,
	Hread,
	Hend,
	Hwrite,
	Herr,

	Hsize = HBufSize
};

struct Hio {
	Hio		*hh;			/* next lower layer Hio, or nil if reads from fd */
	int		fd;			/* associated file descriptor */
	ulong		seek;			/* of start */
	uchar		state;			/* state of the file */
	uchar		xferenc;		/* chunked transfer encoding state */
	uchar		*pos;			/* current position in the buffer */
	uchar		*stop;			/* last character active in the buffer */
	uchar		*start;			/* start of data buffer */
	ulong		bodylen;		/* remaining length of message body */
	uchar		buf[Hsize+32];
};

/*
 * request line
 */
/*	RFC says URI format is:
*		scheme://netlocation/path;params?query#fragment
*	params is:
*		param;param;..;param
*	where vocabularies are taken from RFC
*	Note that scheme info is not in request headers,
*	we can get the info from port no.
*	-Kenar-	*/
struct HttpReq
{
	/*
	 * request line
	 */
	char		*urihost;	//	netlocation,	host in "http://host/...."
	char		*uripath;		//	/path part on URI
	char		*user;	//	alice in "/~alice/...."
	char		*meth;	//	GET, POST, HEAD, ...
	char		*path;	// 	path in httpd space -Kenar-
	char		*intpath; // internal effective path -Kenar-
	char		*params;	//	params	-Kenar-
	char		*options;	//	;params?query	-Kenar-
	char		*ouri;		//	original uri: /path;params?query with %HH -Kenar-
	char		*query;	//	query with %HH
	int		vermaj;
	int		vermin;
};

/*
 * header lines
 */
struct HttpHead
{
	int		closeit;		/* http1.1 close connection after this request? */
	uchar		persist;		/* http/1.1 requests a persistent connection */

	uchar		expectcont;		/* expect a 100-continue */
	uchar		expectother;		/* expect anything else; should reject with ExpectFail */
	ulong		contlen;		/* if != ~0UL, length of included message body */
	HFields		*transenc;		/* if present, encoding of included message body */
	char		*client;
	char		*host;
	HContent	*okencode;
	HContent	*oklang;
	HContent	*oktype;
	HContent	*okchar;
	ulong		ifmodsince;
	ulong		ifunmodsince;
	ulong		ifrangedate;
	HETag		*ifmatch;
	HETag		*ifnomatch;
	HETag		*ifrangeetag;
	HRange		*range;
	char		*conttype; // Kenar
	char		*nextnonce; // Kenar

	/* 	authorization info.
	*	fields that are related to digest authentication are introduced by Kenar
	*	-Kenar- */
	char		*authtype;	/* "", "basic" or "digest" */
	char		*authuser;		/* user for both basic and digest authentication */
	char		*authpass;		/* password in case of basic authentication */
	char		*authrealm;		/* realm */
	char		*authnonce;		/* nonce */
	char		*authresponse;	/* response */
	char		*authuri;	/* uri */
	char		*authdigest;	/* entity-digest */
	char		*authopaque;		/* opaque */
	char		*authalgorithm;		/* algorithm, default MD5. */
		/* authalgorithm is	probably unused.
		*  because MD5 is the only one that is described in RFC2069 */
	char		*authcnonce;		/* cnonce RFC2617 */
	char		*authnc;		/* nonce count RFC2617; it's better to keep as it is. */
	char		*authqop;		/* message quolity of protection RFC2617 */

	/*
	 * experimental headers
	 */
	int		fresh_thresh;
	int		fresh_have;
};

/*
 * all of the state for a particular connection
 */
struct HConnect
{
	void		*private;		/* for the library clients */
	HttpReq		req;
	HttpHead	head;

	Bin		*bin;

	ulong		reqtime;		/* time at start of request */
	char		xferbuf[HBufSize];	/* buffer for making up or transferring data */
	uchar		header[HBufSize + 2];	/* room for \n\0 */
	uchar		*hpos;
	uchar		*hstop;
	Hio		hin;
	Hio		hout;
};

struct Handle
{
	char *mimetype; // assined but currently unused
	char **argv;
	int	hctl;
	char *scriptname;
	char *pathinfo;
};


/*
 * configuration for all connections within the server
 */
//extern	char*		hmydomain;
extern	char*		hversion;
extern	Htmlesc		htmlesc[];
extern	char*	docroot;
extern	char*	port;

/*
 * .+2,/^$/ | sort -bd +1
 */
Hio			*hbodypush(Hio *hh, ulong len, HFields *te);
int			hbuflen(Hio *h, void *p);
int			hcheckcontent(HContent*, HContent*, char*, int);
void			hclose(Hio*);
ulong			hdate2sec(char*);
int			hdatefmt(Fmt*);
void		hbind(HConnect *c, char *name, char *old, int flag);
int			hfail(HConnect*, int, ...);
int			hfail0(HConnect *c, char *error, char *title, char *message);
int			hflush(Hio*);
int			hgetc(Hio*);
int			hgethead(HConnect *c, int many);
int			hinit(Hio*, int, int);
int			hiserror(Hio *h);
int			hload(Hio*, char*);
char			*hlower(char*);
HContent		*hmkcontent(HConnect *c, char *generic, char *specific, HContent *next);
HFields			*hmkhfields(HConnect *c, char *s, HSPairs *p, HFields *next);
char			*hmkmimeboundary(HConnect *c);
HSPairs			*hmkspairs(HConnect *c, char *s, char *t, HSPairs *next);
int			hmoved(HConnect *c, char *uri, char *options);
int			hokmsg(HConnect *c, char *msg);
int			hparseheaders(HConnect*, int timeout);
HSPairs			*hparsequery(HConnect *c, char *query);
int			hparsereq(HConnect *c, int timeout);
int			hprint(Hio*, char*, ...);
int			hputc(Hio*, int);
void			*hreadbuf(Hio *h, void *vsave);
int			hredirected(HConnect *c, char *how, char *uri, char *options);
int			hredir(HConnect *c, char *how, char *uri, char *options);
void			hreqcleanup(HConnect *c);
HFields			*hrevhfields(HFields *hf);
HSPairs			*hrevspairs(HSPairs *sp);
int			http11(HConnect*);
int			httpfmt(Fmt*);
char			*httpunesc(HConnect *c, char *s);
int			hunallowed(HConnect *, char *allowed);
int			hungetc(Hio *h);
char			*hunload(Hio*);
int			hurlfmt(Fmt*);
char			*hurlunesc(HConnect *c, char *s);
int			hwrite(Hio*, void*, int);
int			hxferenc(Hio*, int);
int 		qs2env(HConnect *c, char *qs);

int	uriloop(HConnect *c, char *pat, char *rep, char *huri);
char*	redirect(HConnect *c, char *file, char *path, char **pwebroot, char *scheme, char **hostrootp);
int		acceptpost(HConnect *c, int maxpost);
int		execute(HConnect *c, char **argv, int hctl);
int		ipchk(HConnect *c, char *file, char *ip, char *path);
Handle*		handler(HConnect *c, char *file, char *path);

#pragma			varargck	argpos	hprint	2

/*
 * D is httpd format date conversion
 * U is url escape convertsion
 * H is html escape conversion
 */
#pragma	varargck	type	"D"	long
#pragma	varargck	type	"D"	ulong
#pragma	varargck	type	"U"	char*
#pragma	varargck	type	"H"	char*
