#include <u.h>
#include <libc.h>
#include <bio.h>

#include "httpd.h"
#include "httpsrv.h"
#include "misc.h"



/*
*	The format of "$web/etc/allow" is shown using example:
*	/lecture/our/students
*		192.168.[1-8].*
*		...
*	/lecture/other/students
*		...
*	where ip is given by shell style pattern matching.
*
*	restrictions := restriction restriction ...
*	restriction := pathline iplines
*	pathline := path
*	iplines := ipline ipline ..
*	ipline := spaces ippattern ippattern ..
*
*	# return values
*	0: the IP is allowed
*	-1: the IP is disallowed
*/
int
ipchk(HConnect *c, char *file, char *ip, char *path)
{
	char *p, *line, *s;
	int ok, check;//, m;
	Biobuf *b;
	char word[64];

	//logit(c,"ipchk: ip=%s", ip);

	b = Bopen(file, OREAD);
	if(b){
		ok = 1;
		check = 0;
		while((line = Brdline(b, '\n')) != nil){
			line[Blinelen(b)-1] = 0;
			s = strchr(line, '#');
			if(s != nil)
				*s = '\0'; 	/* chop comment */
			/* skip null line */
			if(*skip(line," \t") == 0)
				continue;

			if(strchr(" \t", line[0])){	// ip pattern of lpath
				if(check == 0) // skip the ip pattern
					continue;
				if(ok)
					continue;
				/* 	how can I get single pattern ?
				*	ip pattern has no space	OK */
				while(line = getword(line, word, sizeof word)){ // asign =
					//logit(c,"ipchk: word=%s", word);
					if(word[0] == '~' && match(ip, word+1) == 0){
						ok = 1;
						break;
					}
					if(match(ip, word)){
						ok = 1;
						break;
					}
				}
				if(!ok) // examine more ip line
					continue;
			} else {	// path
				/* first we must check previous analysys */
				if(check && !ok)
					break;
				/* start new path line */
				p = strpbrk(line," \t");
				if(p)
					*p = 0;
				/* 	check some human error, in error put some
				*	guards and let that know. */
				if(line[0] != '/'){
					logit(c, "warning: path in /etc/allow must begin with '/': %s",line);
					path++;
				}
				if(deeper(path, line)){
					//logit(c, "deeper: %s %s",path,line);
					/* 	then we must confirm
					*	the client IP is allowed to see our data */
					check = 1;
					ok = 0;
				}
				else{
					check = 0;
					ok = 1;
				}
			}
		}
		Bterm(b);
		return  ok - 1;
	}
	return 0;
}
