#include <u.h>
#include <libc.h>
#include "httpd.h"
#include "misc.h"

/*
*	a simple data base
*	the usage is ...
*
*	Db *d = nil;
*	db_put(&d,"alice","18");
*	db_put(&d,"bob","20");
*	print("%s\n",db_get(d,"alice"));
*	print("%s\n",db_get(d,"bob"));
*	print("%s\n",db_get(d,"carol"));
*/

int
db_put(Db **dbp, char *key, char *val)
{	/* the value may be nil */

	Db *p,*q;
	p = (Db*)calloc(sizeof(Db),1);
	if(p == nil)
		return -1;
	p->key = strdup(key);
	if(val != nil)
		p->val = strdup(val);
	p->next = nil;
	if(*dbp == nil){
		*dbp = p;
		return 0;
	}
	q = *dbp;
	while(q->next){
		if(strcmp(q->key, key) == 0){
			free(q->val);
			q->val = p->val;
			return 0;
		}
		q = q->next;
	}
	q->next = p;
	return 0;
}

int
db_add(Db **dbp, char *key, char *val)
{	/* the value may be nil */

	Db *p,*q;
	p = (Db*)calloc(sizeof(Db),1);
	if(p == nil)
		return -1;
	p->key = strdup(key);
	if(val != nil)
		p->val = strdup(val);
	p->next = nil;
	if(*dbp == nil){
		*dbp = p;
		return 0;
	}
	q = *dbp;
	while(q->next)
		q = q->next;
	q->next = p;
	return 0;
}

char*
db_get(Db *db, char *key)
{
	while(db){
		if(strcmp(db->key, key) == 0)
			return db->val;
		db = db->next;
	}
	return nil;
}

void
db_del(Db **dbp, char *key)
{
	Db *db, *db0, *p;

	if(dbp == nil || *dbp == nil)
		return;
	db = *dbp;
	/* key matching at the first is very special */
	if(strcmp(db->key, key) == 0){
		*dbp = db->next;
		free(db->key);
		free(db->val);
		free(db);
		return;
	}

	p = db0 = db;
	while(db){
		if(strcmp(db->key, key) == 0){
			p->next = db->next;
			free(db->key);
			free(db->val);
			free(db);
			return;
		}
		p = db;
		db = db->next;
	}
	*dbp = db0;
	return;
}

void
db_free(Db *db)
{
	Db *p;
	while(db){
		free(db->key);
		free(db->val);
		/* 	codes here is unreliable but is practically possible.
		*	However I avoided
		*		free(db);
		*		db = db->next;
		*	because free(db) might put something at db.
		*	You will find similar codes aplied here in contentinit().
		*	-Kenar- */
		p = db->next;
		free(db);
		db = p;
	}
}

//#define TEST
#ifdef TEST
void
db_print(Db *db)
{
	while(db){
		print("%s %s\n", db->key, db->val);
		db = db->next;
	}
}

void
usage(void)
{
	fprint(2,"usage: 8.out\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	Db *d = nil;
	Db *p;
	char buf[80];
	
	db_put(&d,"alice","18");
	db_put(&d,"bob","20");
	db_print(d);

	print("%s\n",db_get(d,"alice"));
	print("%s\n",db_get(d,"bob"));
	print("%s\n",db_get(d,"carol"));

	print("\n");
	db_put(&d,"alice","23");
	db_print(d);


	print("\n");
	db_put(&d,"alice","25");
	db_print(d);

	print("\n");

	db_del(&d, "bob");
	db_print(d);
	db_del(&d, "bob");
	db_print(d);

	db_free(d);
}
#endif
