#include <u.h>
#include <libc.h>
#include <bin.h>
#include "httpd.h"
#include "httpsrv.h"
#include "misc.h"

typedef struct Error	Error;

struct Error
{
	char	*num;
	char	*concise;
	char	*verbose;
};

Error errormsg[] =
{
	[HInternal]	{"500 Internal Error", "Internal Error",
		"This server could not process your request due to an internal error."},
	[HCGIErr]	{"500 Internal Error", "Internal Error",
		"This server could not process your request due to an error: %s"},
	[HTempFail]	{"500 Internal Error", "Temporary Failure",
		"The object %s is currently inaccessible.<p>Please try again later."},
	[HUnimp]	{"501 Not Implemented", "Command not implemented",
		"This server does not implement the %s command."},
	[HUnkVers]	{"501 Not Implemented", "Unknown http version",
		"This server does not know how to respond to http version %s."},
	[HBadCont]	{"501 Not Implemented", "Impossible format",
		"This server cannot produce %s in any of the formats your client accepts."},
	[HCgiTimeout] {"504 Gateway Timeout", "Gateway Timeout", 
		"This server could not process your request due to timeout"},
	[HBadReq]	{"400 Bad Request", "Strange Request",
		"Your client sent a query that this server could not understand."},
	[HSyntax]	{"400 Bad Request", "Garbled Syntax",
		"Your client sent a query with incoherent syntax."},
	[HBadSearch]	{"400 Bad Request", "Inapplicable Search",
		"Your client sent a search that cannot be applied to %s."},
	[HUnauth] {"401 Unauthorized", "Authentication required",
		"You are not autherized to see the data %s."}, //Kenar
	[HNoSearch]	{"403 Forbidden", "Search not supported",
		"The object %s does not support the search command."},
	[HNoData]	{"403 Forbidden", "No data supplied",
		"Search or forms data must be supplied to %s."},
	[HExpectFail]	{"403 Expectation Failed", "Expectation Failed",
		"This server does not support some of your request's expectations."},
	[HForbidden]	{"403 Forbidden", "Forbidden",
		"You are not allowed to see the object %s."},
	[HNotFound]	{"404 Not Found", "Object not found",
		"The object %s does not exist on this server."},
	[HNotAccept] {"406 Not Acceptable", "Not Acceptable",
		"Your Request is not acceptable"}, // Kenar
	[HPrecond] {"412 Precondition Failed", "Precondition Failed",
		"The entity requested does not match the existing entity."}, // Kenar
	[HReqSize]	{"413 Request Entity Too Large", "Request Entity Too Large",
		"Post data size over limit %d, not acceptable."}, // -Kenar-
	[HReqRange] {"416 Request range not satisfiable.", "Bad Request Range",
		"No bytes are avaible for the range you requested.\n"},  // -Kenar-
	[HUnavailable]	{"503 Service Unavailable", "Service Unavailable",
		"Now this service is unavailable."}, // -Kenar-
	[HOK]		{"200 OK", "everything is fine"},
};

	char	*num;
	char	*concise;
	char	*verbose;


int
hfail0(HConnect *c, char *error, char *title, char *message)
{
	Hio *hout;
	int n;
	char *s;
	Db *db;

	if(title == nil){
		s = strchr(error,' ');
		if(s == nil)
			return hfail(c, HInternal);
		title = ++s;
	}

	if(message == nil)
		message = title;
	if(strncmp(message,"<html>", 6) == 0 || strncmp(message,"<!DOCTYPE ", 10) == 10)
		n = snprint(c->xferbuf, HBufSize, "%s", message);
	else
		n = snprint(c->xferbuf, HBufSize,
			"<html><head><title>%s</title></head>\n<body><h1>%s</h1>\n%s</body></html>\n",
			title, title, message);
	hout = &c->hout;
	hprint(hout, "%s %s\r\n", hversion, error);
	hprint(hout, "Date: %D\r\n", time(nil));
	hprint(hout, "Server: %s\r\n", serversoft);
	hprint(hout, "Content-Type: text/html\r\n");
	hprint(hout, "Content-Length: %d\r\n", n);

	db = msgh;
	while(db){
		hprint(hout, "%s %s\r\n", db->key, db->val);
		db = db->next;
	}

	if(c->head.closeit)
		hprint(hout, "Connection: close\r\n");
	else if(!http11(c))
		hprint(hout, "Connection: Keep-Alive\r\n");
	hprint(hout, "\r\n");

	if(c->req.meth == nil || strcmp(c->req.meth, "HEAD") != 0)
		hwrite(hout, c->xferbuf, n);

	logit(c, "Reply: %s", error);
	return hflush(hout);
}

int
hfail(HConnect *c, int reason, ...)
{
	char makeup[HBufSize];
	va_list arg;
	int n;

	va_start(arg, reason);
	vseprint(makeup, makeup+HBufSize, errormsg[reason].verbose, arg);
	va_end(arg);

	return hfail0(c, errormsg[reason].num, errormsg[reason].concise, makeup);
}
