#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ndb.h>
#include <ctype.h>
#include "httpd.h"
#include "httpsrv.h"
#include "misc.h"

extern int lockouttime;
static void nsop(int argc, char *argv[]);

/*	a debugging tool	*/
void
nsdump(char *ns)
{
	Dir *d;
	int fd;
	long m,n;
	fd = open(ns, OREAD);
	m = dirreadall(fd,&d);
	for(n=0;n<m; n++)
		syslog(1,"http", "nsdump: %s/%s", ns, d[n].name);
	free(d);
	close(fd);
}




void*
ezalloc(ulong n)
{
	void *p;

	p = malloc(n);
	if(p == nil)
		sysfatal("out of memory");
	memset(p, 0, n);
	return p;
}

char*
estrdup(char *s)
{
	s = strdup(s);
	if(s == nil)
		sysfatal("out of memory");
	return s;
}


int
xstrcmp(char *s, char *t)
{
	if(s == nil){
		if(t)
			return -1;
		return 0;
	}
	if(t)
		return strcmp(s,t);
	return 1;
}

char*
trim(char *s)
{
	char *t;
	while(isspace(*s))
		s++;
	t = strchr(s, 0);
	t--;
	while(isspace(*t ))
		t--;
	t++;
	*t = 0;
	return s;
}

char *
strlower(char *s)
{
	char *s0;
	s0 = s;
	while(*s){
		*s = tolower(*s);
		s++;
	}
	return s0;
}

/*
*	test if the path is deeper then or equal to ref
*	deeper(path, ref) returns matched size
*		0 if path is not deeper
*		matched size if deeper # the size is used in basic_auth()
*	example:
*	# path	ref
*	/abc		/abc			# return 4
*	/abc		/abc/		# return 4
*	/abc/	/abc			# return 4
*	/abc/	/abc/		# return 4
*	/abc		/abcdef		# return 0
*	/abc		/abc/def		# return 0
*	/abc/	/abcdef		# return 0
*	/abc/def	/abc/		# return 4
*	/abc/def	/abc			# return 4
*	/abc		/			# return 1
*	a		a			# return 1
*	a/		a			# return 1
*	a/		a/			# return 1
*
*	path is:
*	-	absolute path
*	-	relative path
*	-	canonical uri of the form: scheme://host:port/path
*/
int
deeper(char *path, char *ref)
{
	int n;
	int m = 0;
	int f = 0;
	n = strlen(ref);
	if(n == 1 && ref[0] == '/' && path[0] == '/')
		return 1;
	if(ref[n-1] == '/'){
		ref[n-1] = 0;
		n--;
		f = 1;
	}
	if(strncmp(path, ref, n) == 0 && (path[n] == 0 || path[n] == '/'))
		m = n;
	if(f)
		ref[n] = '/';
	return m;
}


#ifdef NOTE
*	shell style path matching function
*	I modified to simplify original one
*	and made some modification of the matching rule.
*	now, '/' is a simplly regular charactor except
*	"/*/" is matched to "/".
*	ref: /sys/src/cmd/rc/glob.c
*	-Kenar-
*
 * 	Does the string s match the pattern p
 * 	* matches any sequence of characters
 * 	? matches any single character
 * 	[...] matches the enclosed list of characters
 *	~ in the beginning of [...] means compliment
 *	- in [...] means range in ascii order
 *
 *	return value
 *	1: matched
 *	0: not matched
 #endif
int
match(char *s, char *p)
{
	char *s0;
	int compl, hit, lo, hi, t, c;
	s0 = s;
	for(;*p; s++,p++){
		switch(*p){
		case '*':
			++p;
			if(*p == '/' && s > s0)
				if(match(s-1, p)) return 1;
			for(;;){
				if(match(s, p)) return 1;
				if(!*s) break;
				s++;
			}
			return 0;
		case '?':
			if(*s=='\0') return 0;
			break;
		case '[':
			if(*s=='\0') return 0;
			c=*s;
			p++;
			compl=*p=='~';
			if(compl) p++;
			hit=0;
			while(*p!=']'){
				if(*p=='\0') return 0;		/* syntax error */
				lo=*p;
				p++;
				if(*p!='-') hi=lo;
				else{
					p++;
					if(*p=='\0') return 0;	/* syntax error */
					hi=*p;
					p++;
					if(hi<lo){ t=lo; lo=hi; hi=t; }
				}
				if(lo<=c && c<=hi) hit=1;
			}
			if(compl) hit=!hit;
			if(!hit) return 0;
			break;
		default:
			if(*p != *s) return 0;
		}
	}
	return *s=='\0';
}


char *
skip(char *s, char *p)
{
	while(*s && strchr(p, *s))
		s++;
	return s;
}

char *
skipto(char *s, char *p)
{
	while(*s && !strchr(p, *s))
		s++;
	return s;
}

char *
skipapha(char *s)
{
	while(*s && isalpha(*s))
		s++;
	return s;
}


/*	getword get a word from s and returns point
*	next to search ( the point is next char of the word)
*	if there is no word, it returns nil
*	-Kenar-	*/
char *
getword(char *s, char *w, int size)
{
	char *r;
	r = w + size;
	s = skip(s, " \t");
	if(*s == 0)
		return nil;
	while(*s && !strchr(" \t", *s)){
		if(w == r){
			w--;
			break;
		}
		*w++ = *s++;
	}
	*w = 0;
	return s;
}

/*
*	Pegasus before 2.2
*	executable returns 1
*	if the file is executable for `others'
*	else returns 0
*
*	Pegasus 2.2
*	executable returns 1
*	if the file is executable for the httpd
*	else returns 0
*/
int
executable(char *path)
{
/*	// Pegasus before 2.2
	Dir *dir;
	dir = dirstat(path);
	if(dir == nil)
		return 0;
	if(dir->mode & DMEXEC)
		return 1;
	free(dir);
	return 0;
*/
	return !access(path, AEXEC);
}

int
waitfor(int pid, char *msg)
{	int cpid;
	Waitmsg *w;
	
	while((w = wait()) != nil){
		if(pid == w->pid){
			strncpy(msg, w->msg, ERRMAX);
			free(w);
			return 0;
		}
		free(w);
	}
	return -1;
}

/*	bind only addns
*	original source: /sys/src/libauth/newns.c	*/
#define NARG	5
int
chgns(char *file)
{
	Biobuf *spec;
	int argc;
	char *cmd, *argv[NARG];

	spec = Bopen(file, OREAD);
	if(spec == 0){
		werrstr("can't open %s: %r", file);
		return -1;
	}
	while(cmd = Brdline(spec, '\n')){
		cmd[Blinelen(spec)-1] = '\0';
		while(*cmd==' ' || *cmd=='\t')
			cmd++;
		if(*cmd == '#')
			continue;
		/* we need not support '#' tree	*/
		argc = getfields(cmd, argv, NARG, 1, " \t");
		if(argc)
			nsop(argc, argv);
	}
	Bterm(spec);
	return 0;
}

/* 	bind only nsop
*	original source: /sys/src/libauth/newns.c	*/
static void
nsop(int argc, char *argv[])
{
	char *argv0;
	ulong flags;

	flags = 0;
	argv0 = 0;
	ARGBEGIN{
	case 'a':
		flags |= MAFTER;
		break;
	case 'b':
		flags |= MBEFORE;
		break;
	case 'c':
		flags |= MCREATE;
		break;
	case 'C':
		// flags |= MCACHE;
		break;
	}ARGEND

	if(!(flags & (MAFTER|MBEFORE)))
		flags |= MREPL;
	if(strcmp(argv0, "bind") == 0 && argc == 2 && *argv[0] != '#')
		bind(argv[0], argv[1], flags);
}

static char *
cat3(char *buf, char *s1, char *s2, char *s3)
{
	strcpy(buf, s1); strcat(buf, s2); strcat(buf, s3);
	return buf;
}


/*	stolen from /sys/src/cmd/aux/listen1.c	*/
char*
netaddr(char *dir, char *which)
{
	static char buf[128];
	char *p;
	int n, fd;
	/*	which: "loacl","remote"	*/
	snprint(buf, sizeof buf, "%s/%s", dir, which);
	fd = open(buf, OREAD);
	if(fd < 0)
		return "";
	n = read(fd, buf, sizeof(buf));
	close(fd);
	if(n > 0){
		buf[n] = 0;
		p = strchr(buf, '!');
		if(p)
			*p = 0;
		return buf;
	}
	return "";
}


/*
*	blacklist
*/
#define NBLIST	10
typedef struct BlackIP BlackIP;
struct BlackIP
{
	char *ip;
	long time; // registered time
};

static BlackIP blacklist[NBLIST];
static int nblist;
/*	examine if the ip is in blacklist	*/
int
black(char *ip)
{
	int n, k;
	long now;
	now = time(nil);
	n = nblist - NBLIST;
	if(n < 0)
		n = 0;
	for(k = nblist -1; k >= n; k--)
		if(strcmp(blacklist[k % NBLIST].ip, ip) == 0
		&& blacklist[k % NBLIST].time + lockouttime > now)
			return 1;
	return 0;
}

void
putblist(char *ip)
{
	int n;
	n = nblist % NBLIST;
	if(blacklist[n].ip){
		free(blacklist[n].ip);
		blacklist[n].ip = nil;
	}
	blacklist[n].ip = estrdup(ip);
	blacklist[n].time = time(nil);
	nblist++;
}


/*
*	httpd configuration
*/

static Db *configlist;

int
readconfig(char *file)
{
	char *line, *s;
	Biobuf *b;
	char *field[3];

	b = Bopen(file, OREAD);
	if(b == nil)
		return 0; // need not exist
	while((line = Brdline(b, '\n')) != nil){
		line[Blinelen(b)-1] = 0;
		s = strchr(line, '#');
		if(s != nil)
			*s = '\0'; 	/* chop comment */
		if(getfields(line, field, 3, 1, " \t") < 2)
			continue;
		if(db_put(&configlist, field[0], field[1]) < 0)
			return -1;
	}
	Bterm(b);
	return 0;
}

void
setintconfig(int *var, char *name)
{
	char *p;
	p = db_get(configlist,name);
	if(p)
		*var = atoi(p);
}

void
setstrconfig(char **var, char *name)
{
	char *p;
	p = db_get(configlist,name);
	if(p)
		*var = p;
}

/*	I hate interface like:
*	hcat(c, arg1, arg2, nil)
*	because I forget last nil	*/
char *
hcat2(HConnect *c, char *a1, char *a2)
{
	int m;
	char *b;
	m = strlen(a1) + strlen(a2) + 1 + UTFmax;
	b = halloc(c, m);
	strcpy(b, a1);
	strcat(b, a2);
	return b;
}

/*	I hate interface like:
*	hcat(c, arg1, arg2, nil)
*	because I forget last nil	*/
char *
hcat3(HConnect *c, char *a1, char *a2, char *a3)
{
	int m;
	char *b;
	m = strlen(a1) + strlen(a2) + 1 + UTFmax;
	b = halloc(c, m);
	strcpy(b, a1);
	strcat(b, a2);
	strcat(b, a3);
	return b;
}

