#include <u.h>
#include <libc.h>
#include <bin.h>
#include "httpd.h"
#include "httpsrv.h"

typedef struct Strings		Strings;

struct Strings
{
	char	*s1;
	char	*s2;
};

static	char*		abspath(HConnect *cc, char *origpath, char *curdir);
static	int		getc(HConnect*);
static	char*		hgetword(HConnect*);
static	Strings		parseuri(HConnect *c, char*);
static	Strings		stripmagic(char*);
static 	Strings		split(char *uri,char sep);

/*
 * parse the next request line
 * returns:
 *	1 ok
 *	0 eof
 *	-1 error
 */
int
hparsereq(HConnect *c, int timeout)
{
	Strings ss;
	char *vs, *v, *query, *uri, *origuri, *extra, *params, *options;
	int m;

	if(c->bin != nil){
		logit(c, "hparsereq: c->bin != nil");
		hfail(c, HInternal);
		return -1;
	}

	/*
	 * serve requests until a magic request.
	 * later requests have to come quickly.
	 * only works for http/1.1 or later.
	 */
	alarm(timeout);
	if(hgethead(c, 0) < 0)
		return 0;
	alarm(0);

	/* DEBUG
	m = c->hstop - c->hpos;
	v = halloc(c,m + 1);
	memcpy(v, c->hpos, m);
	v[m] = 0;
	logit(c,"DEBUG: %s", v); // GET /foo HTTP/1.1
	*/

	c->reqtime = time(nil);
	c->req.meth = hgetword(c);
	if(c->req.meth == nil){
		hfail(c, HSyntax);
		return -1;
	}
	uri = hgetword(c);
	if(uri == nil || strlen(uri) == 0){
		hfail(c, HSyntax);
		return -1;
	}
	v = hgetword(c);
	//logit(c, "DEBUG: hparsereq: v=%s", v);
	if(v == nil){
#ifdef nodef
		if(strcmp(c->req.meth, "GET") != 0){
			hfail(c, HUnimp, c->req.meth);
			return -1;
		}
#endif
		if(strcmp(c->req.meth, "GET") != 0)
			logit(c,"DEBUG: %s", c->req.meth);
		/*	I don't fully understand the line  -Kenar- */


		c->req.vermaj = 0;
		c->req.vermin = 9;
	}else{
		vs = v;
		if(strncmp(vs, "HTTP/", 5) != 0){
			hfail(c, HUnkVers, vs);
			return -1;
		}
		vs += 5;
		c->req.vermaj = strtoul(vs, &vs, 10);
		if(*vs != '.' || c->req.vermaj != 1){
			hfail(c, HUnkVers, vs);
			return -1;
		}
		vs++;
		c->req.vermin = strtoul(vs, &vs, 10);
		if(*vs != '\0'){
			hfail(c, HUnkVers, vs);
			return -1;
		}

		extra = hgetword(c);
		if(extra != nil){
			hfail(c, HSyntax);
			return -1;
		}
	}

	/*
	*	uri is used below...
	*	we will get uri such that
	*	/foo.html
	*	if client send a request:  http://ar.aichi-u.ac.jp/foo.html
	*	note that HTTP requset is:
	*	GET /foo.html HTTP/1.1
	*	-Kenar-
	*/

	/*
	 * the fragment is not supposed to be sent
	 * strip it 'cause some clients send it
	 */

	origuri = uri;
	uri = strchr(origuri, '#');
	if(uri != nil)
		*uri = 0;
	//logit(c,"DEBUG: ouri=%s", origuri);
	c->req.ouri = hstrdup(c,origuri);


	/*
	 * http/1.1 requires the server to accept absolute
	 * or relative uri's.  convert to relative with an absolute path
	 */
	if(http11(c)){
		ss = parseuri(c, origuri);
		uri = ss.s1;
		c->req.urihost = ss.s2;
		if(uri == nil){
			hfail(c, HBadReq);
			return -1;
		}
		origuri = uri;
	}

	/*	getting option part. option is,
	*	;params?query
	*	-Kenar-	*/
	v = strchr(origuri,';');
	if(v == nil)
		v = strchr(origuri,'?');
	if(v)
		options = hstrdup(c, v);
	else
		options = "";
	c->req.options = options;

	/*
	 * munge uri for search, protection, and magic
	 */

	/*
	*	strip params and query from uri
	*	The input uri is of the form  # RFC2068
	*	/path;params?query
	*	hurlunesc should not be done here.
	*	-Kenar-
	*/

	/*	strip query first and then params. don't change this order	*/
	ss = split(origuri,'?');
	origuri = ss.s1;
	query = ss.s2;
	ss = split(origuri,';');
	origuri = ss.s1;
	params = ss.s2;

	/*	we should apply hurlunesc only to path part */
	uri = hurlunesc(c, origuri);
	uri = abspath(c, uri, "/");
	if(uri == nil || uri[0] == '\0'){
		hfail(c, HBadReq);
		return -1;
	}

	c->req.uripath = uri;
	c->req.params = params;
	c->req.query = query;

	return 1;
}

static Strings
parseuri(HConnect *c, char *uri)
{
	Strings ss;
	char *urihost, *p;

	urihost = nil;
	if(uri[0] != '/'){
		if(cistrncmp(uri, "http://", 7) != 0){
			ss.s1 = nil;
			ss.s2 = nil;
			return ss;
		}
		uri += 5;	/* skip http: */
	}

	/*
	 * anything starting with // is a host name or number
	 * hostnames constists of letters, digits, - and .
	 * for now, just ignore any port given
	 */
	if(uri[0] == '/' && uri[1] == '/'){
		urihost = uri + 2;
		p = strchr(urihost, '/');
		if(p == nil)
			uri = hstrdup(c, "/");
		else{
			uri = hstrdup(c, p);
			*p = '\0';
		}
		p = strchr(urihost, ':');
		if(p != nil)
			*p = '\0';
	}

	if(uri[0] != '/' || uri[1] == '/'){
		ss.s1 = nil;
		ss.s2 = nil;
		return ss;
	}

	ss.s1 = uri;
	ss.s2 = hlower(urihost);
	return ss;
}


/*
 *  to circumscribe the accessible files we have to eliminate ..'s
 *  and resolve all names from the root.
 */
static char*
abspath(HConnect *cc, char *origpath, char *curdir)
{
	char *p, *sp, *path, *work, *rpath;
	int len, n, c;

	if(curdir == nil)
		curdir = "/";
	if(origpath == nil)
		origpath = "";
	work = hstrdup(cc, origpath);
	path = work;

#ifdef nodef
	/*
	* 	remove any really special characters
	*
	*	Is this really necessary ? We should examine. -Kenar-
	*/
	for(sp = "`;| "; *sp; sp++){
		p = strchr(path, *sp);
		if(p)
			*p = 0;
	}
#endif

	len = strlen(curdir) + strlen(path) + 2 + UTFmax;
	if(len < 10)
		len = 10;
	rpath = halloc(cc, len);
	if(*path == '/')
		rpath[0] = 0;
	else
		strcpy(rpath, curdir);
	n = strlen(rpath);

	while(path){
		p = strchr(path, '/');
		if(p)
			*p++ = 0;
		if(strcmp(path, "..") == 0){
			while(n > 1){
				n--;
				c = rpath[n];
				rpath[n] = 0;
				if(c == '/')
					break;
			}
		}else if(strcmp(path, ".") == 0){
			;
		}else if(n == 1)
			n += snprint(rpath+n, len-n, "%s", path);
		else
			n += snprint(rpath+n, len-n, "/%s", path);
		path = p;
	}

	return rpath;
}

static char*
hgetword(HConnect *c)
{
	char *buf;
	int ch, n;

	while((ch = getc(c)) == ' ' || ch == '\t' || ch == '\r')
		;
	if(ch == '\n')
		return nil;
	n = 0;
	buf = halloc(c, 1);
	for(;;){
		switch(ch){
		case ' ':
		case '\t':
		case '\r':
		case '\n':
			buf[n] = '\0';
			return hstrdup(c, buf);
		}

		if(n < HMaxWord-1){
			buf = bingrow(&c->bin, buf, n, n + 1, 0);
			if(buf == nil)
				return nil;
			buf[n++] = ch;
		}
		ch = getc(c);
	}
	return nil;
}

static int
getc(HConnect *c)
{
	if(c->hpos < c->hstop)
		return *c->hpos++;
	return '\n';
}

static Strings
split(char *uri,char sep)
{
	Strings ss;
	char *p;
	p = strchr(uri, sep);
	if(p)
		*p++ = 0;
	ss.s1 = uri;
	ss.s2 = p;
	return ss;
}

