#include <u.h>
#include <libc.h>
#include <bin.h>
#include "httpd.h"
#include "httpsrv.h"
#include "misc.h"


int
hredirected(HConnect *c, char *how, char *uri, char *options)
{
	Hio *hout;
	char *s, *ss, *host, *sh, *path;
	int n, status;

	/*	how: "301 Moved Permanently", for example	*/

	if(options == nil)
		options = "";

	//	logit(c, "DEBUG: hredirected: %U[%s]\n", uri, options);

	/*
	*	uri may be:
	*		scheme://host/path
	*		/path
	*		/~user/path
	*		# Note that we don't handle relative path
	*	and req.uri is
	*		scheme://host/path
	*/

	host = c->head.host; // my host

	sh = uriroot; // my "scheme://host"
	path = "";
	if(uri[0] != '/'){
		/*	we must extract "scheme://host" from the uri	*/
		s = strchr(uri,':');
		/*	what should I do if s==nil ?
		*	then we don't redirect and put the event to log file */
		if(s == nil || s[1] != '/' || s[2] != '/'){
			logit(c, "# illegal redirection to: %s", uri);
			return -1;
		}
		s += 3;
		s = strchr(s,'/');
		if(s){
			*s = 0;
			sh = hstrdup(c, uri);
			*s = '/';
			path = s;
		}
		else
			sh = hstrdup(c, uri);
	}
	else
		path = uri;
	
#ifdef nodef
	if(strchr(uri, ':')){
		/*	then redirects to other scheme or to other host
		*	"scheme://host" is given in the uri		*/
		host = "";
	}
	else if(uri[0] != '/'){
		s = strrchr(c->req.uri, '/');
		if(s != nil)
			*s = '\0';
		ss = halloc(c, strlen(c->req.uri) + strlen(uri) + 2 + UTFmax);
		sprint(ss, "%s/%s", c->req.uri, uri);
		uri = ss;
		if(s != nil)
			*s = '/';
	}
#endif

	n = snprint(c->xferbuf, HBufSize, 
			"<html><head><title>Redirection</title></head>\r\n"
			"<body><h1>Redirection</h1>\r\n"
			"Your selection can be found <a href=\"%U\"> here</a>.<p></body></html>\r\n", uri);

	hout = &c->hout;
	hprint(hout, "%s %s\r\n", hversion, how);
	hprint(hout, "Date: %D\r\n", time(nil));
	hprint(hout, "Server:  %s\r\n", serversoft);
	hprint(hout, "Content-type: text/html\r\n");
	hprint(hout, "Content-Length: %d\r\n", n);
	/*	Location: must be full uri
	*	that is, scheme://host/path
	*	%U format is legal only for path part		*/
	hprint(hout, "Location: %s%U%s\r\n", sh, path,options);

	if(c->head.closeit)
		hprint(hout, "Connection: close\r\n");
	else if(!http11(c))
		hprint(hout, "Connection: Keep-Alive\r\n");
	hprint(hout, "\r\n");

	if(strcmp(c->req.meth, "HEAD") != 0)
		hwrite(hout, c->xferbuf, n);

	//logit(c, "DEBUG: Redirect: %s %s[%U][%s]\n", how, sh, path, options);
	status = hflush(hout);
	if(status)
		logit(c,"# hflush in hredirected");
	return status;
}

/*
*	It's better to close connection if destination is not
*	this name space.
*	uri:
*	"/~who/path"
*	"/path"
*/
int
hredir(HConnect *c, char *how, char *uri, char *options)
{
	char *p, *u;
	u = nil; // user in uri
	if(uri[0] == '/'){
		if(uri[1] == '~'){
			/*	uri == "/~alice/foo" for example */
			p = strchr(uri+2, '/');
			if(p)
				*p = 0;
			u = hstrdup(c, uri+2);
			/*	the u == "alice" */
			if(p)
				*p = '/';
		}
		if(xstrcmp(c->req.user, u) != 0)
			c->head.closeit = 1;
	}
	else
		c->head.closeit = 1;
	
	return hredirected(c, how, uri, options);
}

/*
*	It's better to close connection if destination is not
*	this name space.
*	uri:
*	"/~who/path"
*	"/path"
*/
int
hmoved(HConnect *c, char *uri, char *options)
{
	return hredir(c, "301 Moved Permanently", uri, options);
}

