#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <bin.h>
#include "httpd.h"

/*	original code is
int
hurlfmt(Fmt *f)
{
	char buf[HMaxWord*2];
	Rune r;
	char *s;
	int t;

	s = va_arg(f->args, char*);
	for(t = 0; t < sizeof(buf) - 8; ){
		s += chartorune(&r, s);
		if(r == 0)
			break;
		if(r <= ' ' || r == '%' || r >= Runeself)
			t += snprint(&buf[t], sizeof(buf)-t, "%%%2.2x", r);
		else
			buf[t++] = r;
	}
	buf[t] = 0;
	return fmtstrcpy(f, buf);
}
*/


/*
*	Kenar's alternative
*	hurlfmt is used to install "%U" in the format of print function
*	"%U" encode some unsafe chars in the path name to %HH style.
*	the usage is
*		print("%U",path) or print("http://%s%U?%s",host,path,query) or etc.
*	but not
*		print("%U",uri) where uri is whole uri
*	We have no encoding fomat for query part. the needs is currently limited
*	to redirection. then it is more desirable to use original query part
*	for redirection. therefore I don't support the special needs.
*
*	RFC3986: the reserved chars are only
*      reserved    = gen-delims / sub-delims
*      gen-delims  = ":" / "/" / "?" / "#" / "[" / "]" / "@"
*      sub-delims  = "!" / "$" / "&" / "'" / "(" / ")"
*                  / "*" / "+" / "," / ";" / "="
*   	If data for a URI component would conflict with a reserved
*   	character's purpose as a delimiter, then the conflicting data must be
*   	percent-encoded before the URI is formed.
*	why not " %<>\"" ? and why that include '/' ?	-Kenar-
*
*	RFC2396: 
*      	reserved    = ";" | "/" | "?" | ":" | "@" | "&" | "=" | "+" |
*                    "$" | ","
*		excluded	= control, space, delims, unwise	# excluded from uri
*		delims 		= "<" | ">" | "#" | "%" | <">
*   	unwise      = "{" | "}" | "|" | "\" | "^" | "[" | "]" | "`"
*   	Data corresponding to excluded characters must be escaped in order to
*   	be properly represented within a URI.
*
*	-Kenar-
*/
int
hurlfmt(Fmt *f)
{
	char *b, c;
	char *s, *t, *e;
	int n, status;
	//char *unsafe=" %<>\":?#[]@!$&'()*+,;=";
	/*	unsafe: char set that is safe to encode in %HH,
	*	note that we have redirection such that
	*		/foo	/foo;
	*	that is, deeper path part is added as optional part to uri
	*	we should have analize and separate these uri to
	*	pure path part and new optional part,
	*	but Pegasus 2.3 does not do it.
	*	therefore currently we simply:		*/
	char *unsafe=" :@!&=+$,<>%\"{}|\\^[]()*`'";
	/*	where unsage is (following both RFC2396 and RFC3986)
	*	unsafe=reserved+space+delims+unwise-";?#"
	*
	*	Note that we have redirection case
	*	For example, suppose that
	*		path1 is redirected to path2?query
	*	then only path2 should be encoded
	*/

	s = va_arg(f->args, char*);
	n = 0;
	t = s;
	while(*t){
		if(iscntrl(*t)||strchr(unsafe,*t) || (*t & 0x80))
			n += 3; // %HH encoding
		else
			n++;
		t++;
	}
	n++; // for trailing '\0'

	b = malloc(n);
	t = b;
	e = t + n;
	while(*s){
		if(iscntrl(*s)||strchr(unsafe,*s) || (*s & 0x80))
			t = seprint(t,e,"%%%02X", (uchar)*s);
			/* %HH encoding; upper case is recommended by RFC3987 */
		else
			*t++ = *s;
		s++;
	}
	*t = 0;

	status = fmtstrcpy(f, b);
	free(b);
	return status;
}

#ifdef XXX
/*	debug follows 	*/

void
usage(void)
{
	fprint(2,"usage: 8.out ....\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	ARGBEGIN{
	default: usage();
	}ARGEND

	fmtinstall('U', hurlfmt);
	
	while(*argv)
		print("%U\n", *argv++);
}

/*	the result
term% 8c urlfmt.c
term% 8l urlfmt.8
term% 8.out alice '%' シナリオ '音楽' 'ab%c?de%f'
alice
%25
%E3%82%B7%E3%83%8A%E3%83%AA%E3%82%AA
%E9%9F%B3%E6%A5%BD
ab%25c%3Fde%25f
term% 
*/
#endif
