/*
	Qs2env parses query string and create environment variables.
	The string is given in the argument.
	Format of query string is assumed in BNF:
		element *("&" element)
	where
		element = (value | name "=" value)
	If name is not given, then "" (length of zero string) is assumed
	for the name.
	Values are stored in the environment variables named QS_name.

	Example 1:
	qsparse read a query string:
		family=carrol&name=alice&name=bob
	and create two environment variables named "QS_family" and "QS_name"
	with the values
		("carrol") and ("alice","bob")
	respectively.

	Example 2:
	qsparse read a query string:
		alice+bob&carol
	and create an environment variable named "QS_" with the value
		("alice bob","carol")
*/


#include <u.h>
#include <libc.h>
#include <ctype.h>
#include "httpd.h"
#include "httpsrv.h"

#define NAMELEN 128

//Maxenvsize = 16300;// see /sys/src/9/port/devenv.c

static int	addenv(char *name, char *value);
static char* 	plus2space(char *s);
static char*	shapeup(char *s);

/*
*	the step is:
*	1.	&
*	2.	=
*	3.	+
*	4.	%XX
*
*	return values:
*		0: normal end
*		-1: error in name (left side of "=")
*		> 0 : trancated
*	-Kenar-
*/
int
qs2env(HConnect *c, char *qs)
{
	char **q;
	char *p,*s, *t;
	char buf[NAMELEN];
	int n, i;
	int tranc = 0; // trancation count
	/*	count '&' in the string	*/

	if(qs == nil)
		return 0;
	/*	count "&" before "getfields"	*/
	n = 0;
	for(s = qs; *s; s++)
		if(*s == '&')
			n++;
	q = halloc(c, (n + 1)*sizeof(char*));
	/*	"getfields" overwrites the string, so I must ...	*/
	s = strdup(qs);
	n = getfields(s, q, n+1, 0, "&");

	for(i = 0; i < n; i++){
		/*	then q[i] is attr=value or value */
		p = strchr(q[i], '=');
		if(p){
			*p++ = 0;
			/*	then q[i] is attr and p is value */
			s = q[i];
			/* translate and check the the value before put env */
			s = plus2space(s);
			s = hurlunesc(c, s);
			for(t= s; *t; t++)
				if(!isalnum(*t) && !strchr("_", *t))
					return -1;
		}
		else{
			s = "";
			p = q[i];
		}
		/*	then p is value */
		p = plus2space(p);
		p = hurlunesc(c, p);
		p = shapeup(p);
		snprint(buf, sizeof buf, "QS_%s", s);
		/*	don't use putenv here */
		if(addenv(buf, p) < 0)
			tranc++;
	}
	return tranc;
}

/*	memo
term% a=alice
term% xd -c /env/a
0000000   a  l  i  c  e 00
0000006 
term% a=(alice bob)
term% xd -c /env/a
0000000   a  l  i  c  e 00  b  o  b 00
000000a 
*/
static int
addenv(char *name, char *value)
{
	int fd, n, m;
	char buf[64];
	snprint(buf, sizeof buf, "/env/%s", name);
	fd = open(buf, OWRITE);
	if(fd < 0)
		fd = create(buf, OWRITE, 0666);
	if(fd < 0)
		return -1;
	seek(fd, 0, 2);
	m = strlen(value) + 1;
	n = write(fd, value, m);
	close(fd);
	if(n != m)
		return -1;
	return 0;
}


static char *
plus2space(char *s)
{
	char *p;
	for(p = s; *p; p++)
		if(*p == '+')
			*p = ' ';
	return s;
}

static char*
shapeup(char *s)
{
	char *s0,*t;
	s0 = t = s;
	while(*s){
		if(!iscntrl(*s))
			*t++ = *s;
		s++;
	}
	*t = 0;
	return s0;
}
