#include <u.h>
#include <libc.h>
#include <bio.h>
#include "httpd.h"
#include "httpsrv.h"
#include "misc.h"


/*
*	uriloop(pat, rep, huri) returns:
*	 1 if rep makes loop from the pat
*	 0 if not
*
*	huri is incoming uri of the host:
*		scheme://host[:port]
*	pat is one of the form:
*		scheme://host[:port][/path]
*		/path
*	rep is one of the form:
*		scheme://host[:port][/path][;others]
*		/path[;others]
*	scheme and host of pat and rep is assumed to be lower.
*	port part is assumed to be eliminated if it is:
*		80 for http
*		443 for https
*/

int
uriloop(HConnect *c, char *pat, char *rep, char *huri)
{
	int ch;
	char *p, *q;
	char *user, *huri2, *path2;

	/*	case study A: if rewrite file of alice is:
	*		/foo	/~alice/foo
	*		then we have
	*		(pat, rep, huri)==("/foo", "/~alice/foo", "http://ar")
	*		in accessing with "http://ar/~alice/foo".
	*	case study B: if rewrite file of alice is:
	*		/		https://ar/~alice
	*		then we have
	*		(pat, rep, huri)==("/", "https://ar/~alice/foo", "https://ar")
	*/
	//syslog(0,HTTPLOG,"DEBUG: uriloop: pat=%s rep=%s huri=%s", pat, rep, huri);
	/*	comparison under common form is required	*/
	p = rep;
	if(strncmp(p, "http://", 7) == 0)
		p += 7;
	else if(strncmp(p, "https://", 8) == 0)
		p += 8;
	else if(*rep != '/') // the rep is not to our scheme, not looped
		return 0;
	/* 	then we have two cases for p:
	*	1. *p != '/': p points host part. this case comes if rep begines with scheme
	*	2. *p == '/': p points abosolute path	 */
	ch = 0;
	q = p;
	huri2 = nil;
	if(*p != '/'){
		q = skipto(p, "/");
		ch = *q;
		*q = 0;
		huri2 = hstrdup(c, rep);
		*q = ch;
	}
	path2 = q;
	/*
	*	then
	*		huri2: schem + hosts , i.e., "http://ar" for example
	*	are scheme+host and path in rep
	*	In case A:
	*		(huri2,path2)==("http://ar","/~alice/foo")
	*	In case B:
	*		(huri2,path2)==("https://ar","/~alice/foo")
	*/
	//syslog(0,HTTPLOG,"DEBUG: uriloop: huri2=%s path2=%s", huri2, path2);

	if(huri2 && xstrcmp(huri, huri2) != 0)
		return 0;

	user = c->req.user;
	if(user){
		int n;
		n = strlen(pat) + strlen(user) + 10;
		p = halloc(c,n);
		snprint(p, n, "/~%s%s", user, pat);
		/*	then p is the pattern that includes user part */
		//syslog(0,HTTPLOG,"DEBUG: uriloop: p=%s path2=%s", p, path2);
		/*	In case B:
		*		(p,path2)==("/~alice/","/~arisawa")
		*	How to state these equivalence?
		*/
		if(((n=strlen(p))>0) && p[n-1]=='/')
			p[n-1] = 0;
		if(((n=strlen(path2))>0) && path2[n-1]=='/')
			path2[n-1] = 0;
		/*	side effect to rep!
		*	but we should have done this normalization in earlier stage
		*	probably at reading the rewrite file.	*/
		if(xstrcmp(p,path2) != 0)
			return 0;
		pat = p;
	}

	//logit(c,"DEBUG: uriloop: pat=%s path2=%s", pat, path2);

	if(*path2) // then *path2 == '/'
		return deeper(path2, pat);
	// then pat has not path part, we can asume:
	return (strlen(pat) == 1);
}

#ifdef NOTE
*	path redirection for users and virtual hosts
*	this redirection is used after namespace configuration
*
*	example:
*	request is http://ar/~alice/photo/index.html;args?query
*	then,
*	path is "/photo/index.html"
*	c->req.user is "alice"
*	c->head.host is "ar"
*	
*	How to design rewrite file ?
*	A line consists two fields except comment:
*		pattern	redirection
*
*	We support by this file:
*	1. local redirection
*	2. non local redirection
*	3. port redirection
*	4. scheme redirection
*	5. binding
*	We introduce special charactor '*'
*	then types of second fields are:
*		/path		# local redirection
*		*/path		# binding
*		scheme://host/path		# non local redirection
*		scheme://host:port/path	# non local redirection with port
*	- server must know current scheme and port to prevent redirection loop
*	but how to know current scheme?
*	- we have only two scheme that is accepted by httpd: http and https
*	the latter has certificate.
*	- it is enought to compare host address and port to prevent redirection loop.
*	scheme is needless.
*
*	Example redirection file that can be consulted by both http and https
*		http://vhost	*/vhostwebroot
*		/netlib	ftp://somehost			# note not only http or https
*		/oldplace	/newplace			# local redirection
*
*	Convert second field:
*		http://host	->	http://host:80
*	in checking redirection loop.
*	Pre-assign my scheme host port string such as:
*		http://ar:80
*	Then compare.
*	It is safe to remain scheme part so that we need not talk except http and https.
*
*	How should I do in case:
*		/		http://ar.aichi-u.ac.jp
*		/		http://202.250.160.40
*		/private	https://ar.aichi-u.ac.jp/private
*	These make a redirection loop if httpd in running on ar.aichi-u.ac.jp.
*	Hmm... "//*" is not good idea...
*	Second field requires full address.
*	Example redirection file that can be consulted by both http and https
*		http://vhost	*/vhostwebroot
*		/private	https://ar.aichi-u.ac.jp/private	# this is skipped by https
*		/special	http://ar.aichi-u.ac.jp:8080/special	# this is skipped by 8080 http
*		/netlib	ftp://somehost			# note not only http or https
*		/oldplace	/newplace			# local redirection
*		/		http://ar.aichi-u.ac.jp/	# this is skipped by http
*	
*
*	# return values
*	nil: not redirected
*	else: new uri
*	-Kenar-
#endif

char*
redirect(HConnect *c, char *file, char *path, char **pwebroot, char *scheme, char **phostroot)
{
	char *pat, *rep, *line, *s, *buf;
	Biobuf *b;
	char *field[3];
	int bufsize;
	char huri[64]; // host uri: "http://ar.aichi-u.ac.jp/" ...  # scheme://host[:port]
	b = Bopen(file, OREAD);
	if(b){
		pat = nil;
		rep = nil;
		while((line = Brdline(b, '\n')) != nil){
			line[Blinelen(b)-1] = 0;
			s = strchr(line, '#');
			if(s != nil)
				*s = '\0'; 	/* chop comment */
			if(tokenize(line, field, 2) < 2)
				continue;
			if(deeper(path, field[0])){
				pat = field[0];
				rep = field[1];
				//logit(c, "redirect: deeper: %s %s", pat,rep);
				break;
			}
		}
		if(rep == nil){ // never happen if we have valid rewrite file
			Bterm(b);
			return nil;
		}
		if(strcmp(pat,"/") == 0 && strncmp(path, "/~", 2) == 0){
			*phostroot = hstrdup(c,rep+1);
			Bterm(b);
			return nil;
		}
		if(rep[0] == '*'){
			rep++;
			/*	we cut off trailing '/'	*/
			if(*(rep + strlen(rep) - 1) == '/')
				*(rep + strlen(rep) - 1) = 0;
			//rep = hstrdup(c,rep);	
			if(pwebroot)
				*pwebroot = hstrdup(c,rep);
			#ifdef NOTE
			*	In rewrite file we allow
			*	/car		*/usr/carol/www
			*	/~carol	*/usr/carol/www
			*	/		*/usr/bob/www
			* 	Then path="/car/somepath" should return "/somepath"
			#endif
			/*	cut off trailing "/" for safety.
			*	this is especially true for pat="/"	*/
			if(*(pat + strlen(pat) - 1) == '/')
				*(pat + strlen(pat) - 1) = 0;
			rep = hstrdup(c, path + strlen(pat));
		}
		else {// rep is "/...." or "scheme://...." 
			if((strcmp(scheme,"http") == 0 && strcmp(port,httpport) == 0)
			|| (strcmp(scheme,"https") == 0 && strcmp(port,httpsport) == 0))
				snprint(huri, sizeof huri, "%s://%s", scheme, c->head.host);
			else
				snprint(huri, sizeof huri, "%s://%s:%s", scheme, c->head.host,  port);
			if(uriloop(c, pat, rep, huri)){
				//logit(c, "DEBUG: redirect: uriloop");
				return nil;
			}
			bufsize = strlen(rep) + strlen(path) - strlen(pat) + 1 + UTFmax;
			buf = halloc(c, bufsize);
			s = path + strlen(pat);
			if(*s == '/' || *(rep+strlen(rep) -1) == '/' || *s == 0)
				snprint(buf, bufsize, "%s%s", rep, s);
			else
				snprint(buf, bufsize, "%s/%s", rep, s);
			rep = buf;
			if(pwebroot)
				*pwebroot = nil;
		}

		if(strcmp(pat,"/") == 0)
			*phostroot = rep;
		else{
			while((line = Brdline(b, '\n')) != nil){
				line[Blinelen(b)-1] = 0;
				s = strchr(line, '#');
				if(s != nil)
					*s = '\0'; 	/* chop comment */
				if(tokenize(line, field, 2) < 2)
					continue;
				if(strcmp(pat,"/") == 0)
					break;
			}
			*phostroot = strdup(field[1] + 1); // for example /usr/bob/www
		}
		Bterm(b);
		return rep;
	}
	return nil;
}
