#include <u.h>
#include <libc.h>
#include <auth.h>
#include <bio.h>
#include <libsec.h>
#include "httpd.h"
#include "httpsrv.h"
#include "misc.h"
#include "hauth.h"

extern char *charset;

static void	printtype(Hio *hout, HContent *type, HContent *enc);


/*
 * fd references a file which has been authorized & checked for relocations.
 * send back the headers & it's contents.
 * includes checks for conditional requests & ranges.
 */
int
sendit(HConnect *c, char *file, int hctl)
{
	HRange *r;
	HContents conts;
	HContent *enc=nil;
	HContent *type=nil;
	Hio *hout;
	char *boundary, etag[32];
	int fd;
	Dir *dir;
	ulong tr;
	int n, nw, multir, ok;
	vlong wrote;
	char *s, *p;

	hout = &c->hout;

	/*	content type will be determined as the following order
	*	case CGI
	*	1. Explicit declaration in "Content-Type:" in CGI
	*	2. default value in /etc/handler
	*	case regular file
	*	1. URI, i.e., suffix of the file
	*	2. the content of the requested data
	*	-Kenar-	*/

	s = db_get(cgih, "Content-Type:");
	if(s){
		s = hstrdup(c,s);
		p = strchr(s,'/');
		if(p == nil)
			return hfail(c, HInternal);
		*p = 0;
		type = halloc(c,sizeof(HContent));
		type->generic = s;
		type->specific = ++p;
		db_del(&cgih, "Content-Type:");
	}

	/*	content encoding will be determined as the following order
	*	1. Explicit declaration in "Content-Encoding:" in CGI
	*	2. URI, i.e., suffix of the requested file
	*	3. the content of the requested data
	*	-Kenar-	*/

	s = db_get(cgih, "Content-Encoding:");
	if(s){
		enc = hmkcontent(c, s, nil, nil);
		db_del(&cgih, "Content-Encoding:");
	}

	fd = open(file, OREAD);
	if(fd < 0){
		char buf[256];

		if(strcmp(file,"/tmp/...") == 0){ // Kenar
			logit(c, "# error: /tmp/... not open");
			return hfail(c, HInternal);
		}
		logit(c, "sendit: %s not found", file);
		buf[0] = 0;
		errstr(buf, sizeof buf);
		if(strstr(buf, "permission denied"))
			return hfail(c, HForbidden, file);
		return hfail(c, HNotFound, file);
	}
	if((dir = dirfstat(fd)) == nil){
		close(fd);
		logit(c, "# error: sendit: dirfstat: %s", file);
		return hfail(c, HInternal);
	}

	if(dir->mode & DMDIR)
		return hfail(c, HForbidden, file);

	/*
	 * figure out the type of file and send headers
	 */
	n = -1;
	r = nil;
	multir = 0;
	boundary = nil;
	if(c->req.vermaj == 0 && strcmp(c->req.meth,"GET") != 0)
		logit(c, "# error: inconsistent vermaj");
	if(c->req.vermaj){
		if(type == nil && enc == nil){
			conts = uriclass(c, file);// Kenar
			type = conts.type;
			enc = conts.encoding;
			//logit(c,"DEBUG: uriclass %s/%s %s", type, enc, file);
			if(type == nil && enc == nil){
				n = read(fd, c->xferbuf, HBufSize-1);
				if(n > 0){
					c->xferbuf[n] = '\0';
					conts = dataclass(c, c->xferbuf, n);
					type = conts.type;
					enc = conts.encoding;
				}
			}
		}
		if(type == nil)
			type = hmkcontent(c, "application", "octet-stream", nil);

		snprint(etag, sizeof(etag), "\"%luxv%lux\"", dir->qid.path, dir->qid.vers);
		ok = checkreq(c, type, enc, dir->mtime, etag);
		if(ok <= 0){
			close(fd);
			return ok;
		}

		/*
		 * check for if-range requests
		 */

		//logit(c,"DEBUG: sendit: ifrangedate=%d mtime=%d", c->head.ifrangedate, dir->mtime);

		if(c->head.range == nil
		|| c->head.ifrangeetag != nil && !etagmatch(1, c->head.ifrangeetag, etag)
		|| c->head.ifrangedate != 0 && c->head.ifrangedate < dir->mtime){
			c->head.range = nil;
			c->head.ifrangeetag = nil;
			c->head.ifrangedate = 0;
		}
		/*	Note on the fix:
		*	I think the condition on origial code
		*		c->head.ifrangedate != dir->mtime
		*	is too strong.	-Kenar-	*/

		s = db_get(cgih, "Content-Range:");

		if(s == nil && c->head.range != nil){
			c->head.range = fixrange(c->head.range, dir->length);
			if(c->head.range == nil){

#ifdef nodef
				if(c->head.ifrangeetag == nil && c->head.ifrangedate == 0){
					char buf[64];
					snprint(buf, sizeof(buf), "bytes */%lld", dir->length);
					db_put(&msgh, "Content-Range:", buf);
					hfail(c, HReqRange);
					return 1;
				}
				c->head.ifrangeetag = nil;
				c->head.ifrangedate = 0;
				//logit(c,"DEBUG: sendit: fixrange: nil");
#endif
					char buf[64];
					snprint(buf, sizeof(buf), "bytes */%lld", dir->length);
					db_put(&msgh, "Content-Range:", buf);
					hfail(c, HReqRange);
					//logit(c,"DEBUG: sendit: fixrange: nil");
					return 1;
			}
		}

		s = db_get(cgih, "Status:");
		if(s){
			logit(c, "Status: %s", s);
			hprint(hout, "%s %s\r\n", hversion, s);
			db_del(&cgih,"Status:");
		}
		else{
			if(c->head.range == nil)
				hprint(hout, "%s 200 OK\r\n", hversion);
			else
				hprint(hout, "%s 206 Partial Content\r\n", hversion);
		}

		s = db_get(cgih, "Server:");
		if(s == nil)
			s = serversoft;
		hprint(hout, "Server: %s\r\n", s);
		db_del(&cgih,"Server:");

		s = db_get(cgih, "Date:");
		if(s)
			hprint(hout, "Date: %s\r\n", s);
		else
			hprint(hout, "Date: %D\r\n", time(nil));
		db_del(&cgih,"Date:");

		s = db_get(cgih, "ETag:");
		if(s){
			s = trim(s);
			if(*s != 0)
				hprint(hout, "ETag: %s\r\n", s);
			/* we don't put etag if s == "" */
		}
		else
			hprint(hout, "ETag: %s\r\n", etag);

		db_del(&cgih,"ETag:");

		hprint(hout, "Accept-Ranges: bytes\r\n");

		/*
		 * can't send some entity headers if partially responding
		 * to an if-range: etag request
		 */

		/*	The example is:
				If-Range: Sun, 07 Jan 2007 13:25:23 GMT
				Range: bytes=15154-
		*/

		r = c->head.range;
		if(r == nil){
			hprint(hout, "Content-Length: %lld\r\n", dir->length);
			//logit(c, "DEBUG: Content-Length: %lld\r\n", dir->length);
		}
		else if(r->next == nil){
			s = db_get(cgih, "Content-Range:");
			if(s){
				hprint(hout, "Content-Range: %s\r\n", s);
            	db_del(&cgih,"Content-Range:");
			}
			else{
				hprint(hout, "Content-Length: %ld\r\n", r->stop - r->start + 1);
				hprint(hout, "Content-Range: bytes %ld-%ld/%lld\r\n", r->start, r->stop, dir->length);
				logit(c, "DEBUG: Content-Range:  bytes %ld-%ld/%lld\r\n", r->start, r->stop, dir->length);
			}
		}else{
			multir = 1;
			boundary = hmkmimeboundary(c);
			hprint(hout, "Content-Type: multipart/byteranges; boundary=%s\r\n", boundary);
		}
		if(c->head.ifrangeetag == nil){
			s = db_get(cgih, "Last-Modified:");
			if(s){
				hprint(hout, "Last-Modified: %s\r\n", s);
            	db_del(&cgih,"Last-Modified:");
			}
            else
				hprint(hout, "Last-Modified: %D\r\n", dir->mtime);
			if(!multir)
				printtype(hout, type, enc);
			if(c->head.fresh_thresh)
				hintprint(c, hout, c->req.uripath, c->head.fresh_thresh, c->head.fresh_have);
		}

		if(strcmp(file,"/tmp/...") == 0){ // Kenar
			s = db_get(cgih, "Cache-Control:");
			if(s){
				hprint(hout, "Cache-Control: %s\r\n", s);
            	db_del(&cgih,"Cache-Control:");
			}
			else
				hprint(hout, "Cache-Control: no-cache\r\n");
		}

		s = db_get(cgih, "Connection:");
		if(s){
			hprint(hout, "Connection: %s\r\n", s);
           	db_del(&cgih,"Connection:");
		}
		else{
			if(c->head.closeit)
				hprint(hout, "Connection: close\r\n");
			else if(!http11(c))
				hprint(hout, "Connection: Keep-Alive\r\n");
		}

		/* If hctl == HCPlus then add the unprocessed headers" */
		if(hctl == HCPlus){
			Db *d;
			d = cgih;
			while(d){
				hprint(hout,"%s %s\r\n", d->key, d->val);
				d = d->next;
			}
		}
		hprint(hout, "\r\n");
	}
	if(strcmp(c->req.meth, "HEAD") == 0){
		if(c->head.range == nil)
			logit(c, "Reply: 200 file 0");
		else
			logit(c, "Reply: 206 file 0");
		hflush(hout);
		close(fd);
		return 1;
	}


	if(strcmp(file,"/tmp/...") == 0)
		r = nil;

	/*
	 * send the file if it's a normal file
	 */
	if(r == nil){
		hflush(hout);

		wrote = 0;
        /*	we have already read n bytes of the file
		*	to figure out the type of file	*/	
		if(n > 0)
			wrote = hwrite(hout, c->xferbuf, n);
		if(n <= 0 || wrote == n){
			while((n = read(fd, c->xferbuf, HBufSize)) > 0){
				nw = hwrite(hout, c->xferbuf, n);
				if(nw != n){
					if(nw > 0)
						wrote += nw;
					break;
				}
				wrote += nw;
			}
		}
		close(fd);
		hflush(hout);

		if(dir->length == wrote){
			//logit(c, "DEBUG: sent: %s %lld", file, wrote);
			if(strcmp(file, "/tmp/...") != 0) // Kenar
				logit(c,"sent %s", file);
			return 1;
		}
		logit(c, "Reply: 200 file %lld %lld", dir->length, wrote);
		return -1;
	}

	/*
	 * for multipart/byterange messages,
	 * it is not ok for the boundary string to appear within a message part.
	 * however, it probably doesn't matter, since there are lengths for every part.
	 */
	wrote = 0;
	ok = 1;
	for(; r != nil; r = r->next){
		if(multir){
			hprint(hout, "\r\n--%s\r\n", boundary);
			printtype(hout, type, enc);
			hprint(hout, "Content-Range: bytes %ld-%ld/%lld\r\n", r->start, r->stop, dir->length);
			hprint(hout, "\r\n");
		}
		hflush(hout);

		if(seek(fd, r->start, 0) != r->start){
			ok = -1;
			logit(c,"# sendit: seek error: %s", file);
			break;
		}
		//for(tr = r->stop - r->start + 1; tr; tr -= n){
		for(tr = r->stop - r->start + 1; tr > 0; tr -= n){
			n = tr;
			if(n > HBufSize)
				n = HBufSize;
			if(read(fd, c->xferbuf, n) != n){
				ok = -1;
				logit(c,"# sendit: read error: %s", file);
				goto breakout;
			}

			nw = write(1, c->xferbuf, n);
			//nw = hwrite(hout, c->xferbuf, n);
			if(nw != n){
				if(nw > 0)
					wrote += nw;
				ok = -1;
				logit(c,"# sendit: write error: %s, n=%d nw=%d",file,n,nw);
				logit(c,"# sendit: write error: start=%d, stop=%d",r->start,r->stop);
				goto breakout;
			}
			wrote += nw;
		}
	}
breakout:;
	if(r == nil){
		if(multir){
			hprint(hout, "--%s--\r\n", boundary);
			hflush(hout);
		}
		logit(c, "Reply: 206 partial content %lld/%lld", wrote, dir->length);
	}else
		logit(c, "Reply: 206 early termination %lld/%lld",  wrote, dir->length);
	close(fd);
	return ok;
}

static void
printtype(Hio *hout, HContent *type, HContent *enc)
{
	hprint(hout, "Content-Type: %s/%s", type->generic, type->specific);
	//syslog(1,"http","DEBUG: Content-Type: %s/%s", type->generic, type->specific);

	if((cistrcmp(type->generic, "text") == 0) && charset != nil)
		hprint(hout, ";charset=%s", charset);

	hprint(hout, "\r\n");
	if(enc != nil)
		hprint(hout, "Content-Encoding: %s\r\n", enc->generic);
}

int
etagmatch(int strong, HETag *tags, char *e)
{
	char *s, *t;

	for(; tags != nil; tags = tags->next){
		if(strong && tags->weak)
			continue;
		s = tags->etag;
		if(s[0] == '*' && s[1] == '\0')
			return 1;

		t = e + 1;
		while(*t != '"'){
			if(*s != *t)
				break;
			s++;
			t++;
		}

		if(*s == '\0' && *t == '"')
			return 1;
	}
	return 0;
}

static char *
acceptcont(char *s, char *e, HContent *ok, char *which)
{
	char *sep;

	if(ok == nil)
		return seprint(s, e, "Your browser accepts any %s.<br>\n", which);
	s = seprint(s, e, "Your browser accepts %s: ", which);
	sep = "";
	for(; ok != nil; ok = ok->next){
		if(ok->specific)
			s = seprint(s, e, "%s%s/%s", sep, ok->generic, ok->specific);
		else
			s = seprint(s, e, "%s%s", sep, ok->generic);
		sep = ", ";
	}
	return seprint(s, e, ".<br>\n");
}

/*
 * send back a nice error message if the content is unacceptable
 * to get this message in ie, go to tools, internet options, advanced,
 * and turn off Show Friendly HTTP Error Messages under the Browsing category
 */
static int
notaccept(HConnect *c, HContent *type, HContent *enc, char *which)
{
	char *s, *e;

	e = &c->xferbuf[HBufSize];
	s = c->xferbuf;
	s = seprint(s, e, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\n");
	s = seprint(s, e, "<html>\n<title>Unacceptable %s</title>\n<body>\n", which);
	s = seprint(s, e, "Your browser will not accept this data, %H, because of it's %s.<br>\n", c->req.uripath, which);
	s = seprint(s, e, "It's Content-Type is %s/%s", type->generic, type->specific);
	if(enc != nil)
		s = seprint(s, e, ", and Content-Encoding is %s", enc->generic);
	s = seprint(s, e, ".<br>\n\n");

	s = acceptcont(s, e, c->head.oktype, "Content-Type");
	s = acceptcont(s, e, c->head.okencode, "Content-Encoding");
	s = seprint(s, e, "</body>\n</html>\n");
	return hfail0(c,"406 Not Acceptable", "Not Acceptable", s);
}

/*
 * check time and entity tag conditions.
 */
int
checkreq(HConnect *c, HContent *type, HContent *enc, long mtime, char *etag)
{
	Hio *hout;
	int m;

	hout = &c->hout;
	if(c->req.vermaj >= 1 && c->req.vermin >= 1 && !hcheckcontent(type, c->head.oktype, "Content-Type", 0))
		return notaccept(c, type, enc, "Content-Type");
	if(c->req.vermaj >= 1 && c->req.vermin >= 1 && !hcheckcontent(enc, c->head.okencode, "Content-Encoding", 0))
		return notaccept(c, type, enc, "Content-Encoding");

	/*
	 * can use weak match only with get or head;
	 * this always uses strong matches
	 */
	m = etagmatch(1, c->head.ifnomatch, etag);

	if(m && strcmp(c->req.meth, "GET") != 0 && strcmp(c->req.meth, "HEAD") != 0
	|| c->head.ifunmodsince && c->head.ifunmodsince < mtime
	|| c->head.ifmatch != nil && !etagmatch(1, c->head.ifmatch, etag))
		return hfail(c, HPrecond);

	if(c->head.ifmodsince >= mtime
	&& (m || c->head.ifnomatch == nil)){
		/*
		 * can only send back Date, ETag, Content-Location,
		 * Expires, Cache-Control, and Vary entity-headers
		 */
		hprint(hout, "%s 304 Not Modified\r\n", hversion);
		hprint(hout, "Server: %s\r\n", serversoft);
		hprint(hout, "Date: %D\r\n", time(nil));
		hprint(hout, "ETag: %s\r\n", etag);
		if(c->head.closeit)
			hprint(hout, "Connection: close\r\n");
		else if(!http11(c))
			hprint(hout, "Connection: Keep-Alive\r\n");
		hprint(hout, "\r\n");
		logit(c, "Reply: 304 Not Modified");
		return hflush(hout);
	}
	return 1;
}

/*
 * length is the actual length of the entity requested.
 * discard any range requests which are invalid,
 * ie start after the end, or have stop before start.
 * rewrite suffix requests
 */
HRange*
fixrange(HRange *h, long length)
{
	HRange *r, *rr;

	if(length == 0)
		return nil;

	/*
	 * rewrite each range to reflect the actual length of the file
	 * toss out any invalid ranges
	 */
	rr = nil;
	for(r = h; r != nil; r = r->next){
		if(r->suffix){
			r->start = length - r->stop;
			if(r->start >= length)
				r->start = 0;
			r->stop = length - 1;
			r->suffix = 0;
		}
		if(r->stop >= length)
			r->stop = length - 1;
		if(r->start > r->stop){
			if(rr == nil)
				h = r->next;
			else
				rr->next = r->next;
		}else
			rr = r;
	}

	/*
	 * merge consecutive overlapping or abutting ranges
	 *
	 * not clear from rfc2616 how much merging needs to be done.
	 * this code merges only if a range is adjacent to a later starting,
	 * over overlapping or abutting range.  this allows a client
	 * to request wanted data first, followed by other data.
	 * this may be useful then fetching part of a page, then the adjacent regions.
	 */
	if(h == nil)
		return h;
	r = h;
	for(;;){
		rr = r->next;
		if(rr == nil)
			break;
		if(r->start <= rr->start && r->stop + 1 >= rr->start){
			if(r->stop < rr->stop)
				r->stop = rr->stop;
			r->next = rr->next;
		}else
			r = rr;
	}
	return h;
}


