#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <mp.h>
#include <libsec.h>
#include <auth.h>
#define ERRLEN 256

static char *LOG = "mon";
static char *usage="usage: mon [-d] [-u user] [-r req] cmdpath argument ...";
/*
*	-d: run as daemon mode
*	-u user: run as user
*	-u . : run as current user
*	-r req: factotum format of X.509 certificate signing request 
*/

static char *user=nil;
static int daemon = 0;

static void	execv0(char *argv[]);
static int	waitfor(int pid, char *msg);
static void	becomenone(void);


/*	quoted from su.c	*/
char*
strtrim(char *s)
{
	char *t;
	while(isspace(*s))
		s++;
	t = strchr(s, 0);
	t--;
	while(isspace(*t ))
		t--;
	t++;
	*t = 0;
	return s;
}


/*	quoted from su.c	*/
int
mkcap(char *buf, int n, char *user, char *newuser)
{
	int i, fd;
	char *key;
	int keysize = 10;
	uchar hash[SHA1dlen];

	/* check n >= strlen(usr) + 1 + strlen(newuser) + 1 + 2* keysize +1 */
	if(n < strlen(user) + 1 + strlen(newuser) + 1 + 2* keysize +1){
		werrstr("buf size too small");
		return -1;
	}
	
	/* making a key; key must be a null terminated string */
	srand(truerand());
	for(i = 0; i < keysize; i++)
		sprint(buf+2*i, "%2.2ux", rand());
	key = strdup(buf);

	snprint(buf, n, "%s@%s", user, newuser);
	hmac_sha1((uchar*)buf, strlen(buf), (uchar*)key, strlen(key), hash, nil);

	/*	only hostowner can write to /dev/caphash
	*	it is safe to use #¤/caphash */
	fd = open("#¤/caphash", OWRITE);
	if(fd < 0)
		return -1;
	i = write(fd, hash, sizeof hash);
	close(fd);
	if(i < 0)
		return -1;

	n = snprint(buf, n, "%s@%s@%s", user, newuser, key);
	/*	snprint returns strlen(buf)	*/
	return n;
}

/*	quoted from su.c	*/
int
execute(char *path, char *cmd)
{
	int pid;
	int status;
	int n;
	char *args[32];
	char msg[ERRLEN];

	n = tokenize(cmd, args, 32);
	if(n == 0)
		return 0;
	args[n] = nil;

	switch(pid = fork()) {/* assign = */
	case -1:
		sysfatal("# fork: %r");
	case 0:
		close(0);
		exec(path, args);
		sysfatal("# exec: %r");
	default:
		break;
	}
	/* ensure "/bin/auth/factotum" finished */
	status = waitfor(pid, msg);
	if(status < 0){
		werrstr("waitfor: %r");
		return -1;
	}
	return 0;
}


int
setfactotum(char *key)
{
	int fd;
	int n;
	char *factctl;

	execute("/bin/auth/factotum","factotum -n");

	factctl = "/mnt/factotum/ctl";
	fd = open(factctl, OWRITE);
	if(fd < 0){
		werrstr("open %s: %r", factctl);
		return -1;
	}
	seek(fd, 0, 2);
	n = write(fd, key, strlen(key));
	if(n != strlen(key)){
		werrstr("write %s: %r", factctl);
		return -1;
	}
	close(fd);
	return 0;
}

void
main(int argc, char **argv)
{
	int n, t0, t, pid;
	char buf[1024*4], msg[ERRMAX], *p;
	char *user0; // who invoked
	char *req = nil; // request for certificate
	char *factkey = nil; // factotum for certificate
	int fd;

	ARGBEGIN{
	case 'u':
		user = ARGF();
		break;
	case 'd':
		daemon = 1;
		break;
	case 'r':
		req = ARGF();
		break;
	default:
		sysfatal("# %s", usage);
	}ARGEND

	if(*argv == nil)
		sysfatal("# %s", usage);
	user0 = getuser();
	if(user == nil)
		user = "none";
	if(strcmp(user,".") == 0)
		user = user0;

	if(req){
		fd = open(req, OREAD);
		if(fd < 0)
			sysfatal("# not found %s: %r", req);
		n = read(fd, buf, sizeof buf);
		if( n == sizeof buf)
			sysfatal("# size of %s over %d: %r", req, sizeof buf);
		buf[n] = 0;
		factkey = strdup(buf);
		close(fd);
	}

	if(daemon){
		switch(rfork(RFNOTEG|RFPROC|RFFDG|RFNAMEG)) {
		case -1:
			sysfatal("# fork");
		case 0:
			break;
		default:
			exits(nil);
		}
	}
	/*
	 * open all files we might need before castrating namespace
	 */
	time(nil);
	syslog(0, LOG, nil);
	syslog(0, LOG, "mon start %s by %s run as %s", *argv, user0, user);

	t0 = time(nil);
	for(;;){
		/*	child should be in copied name space	*/
		switch(pid = rfork(RFFDG|RFNOTEG|RFREND|RFPROC|RFNAMEG)){// assign =
		case -1:
			sysfatal("# fork: %r");
		case 0: /* child process */
			if(strcmp(user, "none") == 0)
				becomenone();
			else if(strcmp(user, user0) != 0){
				AuthInfo ai;

				if(mkcap(buf, sizeof buf, user0, user) < 0)
					sysfatal("# mkcap: %r");

				ai.cap = buf;
				ai.cuid = user;
				auth_chuid(&ai, nil);
			}

			if(factkey && setfactotum(factkey) < 0)
				sysfatal("# setfactotum: %r");

			execv0(argv);
			sysfatal("# execv0: %r");
		default: /* parent process */
			syslog(0, LOG, "waiting %d",pid);
			waitfor(pid, msg);
			if(*msg){
				p = strchr(msg,':'); p = strchr(p, ':'); p++;
				p = strtrim(p);
				syslog(0,LOG,"%d: %s", pid, p);
				if(*p == '#')
					sysfatal("# %d:%s", pid, ++p);
			}
			else
				syslog(0,LOG,"terminated");
		}
		t = time(nil);
		if(t -t0 < 5) // something wrong, we exit
			exits(nil);
		t0 = t;
	}
}

static void
becomenone(void)
{
	int fd;

	fd = open("#c/user", OWRITE);
	if(fd < 0 || write(fd, "none", strlen("none")) < 0)
		sysfatal("can't become none");
	close(fd);
	if(newns("none", nil) < 0)
		sysfatal("can't build normal namespace");
}


static void
execv0(char *argv[])
{
	char path[256], wbuf[256],*p;
	if(*argv[0] == '/')
		strcpy(path, argv[0]);
	else if(strncmp(argv[0], "./",2) == 0){
		argv[0]++;
		snprint(path, sizeof path, "%s%s", getwd(wbuf, sizeof wbuf),argv[0]);
	}
	else
		snprint(path, sizeof path, "/bin/%s", argv[0]);
	p = strrchr(path, '/');
	if(p){
		p++;
		argv[0] = p;
	}
	exec(path, argv);
}

static int
waitfor(int pid, char *msg)
{
	Waitmsg *w;
	while((w = wait()) != nil){
		if(w->pid == pid){
			strncpy(msg, w->msg, ERRMAX);
			free(w);
			return 0;
		}
		free(w);
	}
	return -1;
}
